import logging
import os.path
import warnings
from collections import deque
from functools import partial
from pathlib import Path, PurePosixPath as pp

import numpy as np
import h5py
from h5py import File, Group, Dataset, special_dtype, SoftLink, ExternalLink, Reference, RegionReference, check_dtype

from .h5_utils import (BuilderH5ReferenceDataset, BuilderH5RegionDataset, BuilderH5TableDataset, H5DataIO,
                       H5SpecReader, H5SpecWriter, HDF5IODataChunkIteratorQueue)
from ..io import HDMFIO
from ..errors import UnsupportedOperation
from ..warnings import BrokenLinkWarning
from ...build import (Builder, GroupBuilder, DatasetBuilder, LinkBuilder, BuildManager, RegionBuilder,
                      ReferenceBuilder, TypeMap, ObjectMapper)
from ...container import Container
from ...data_utils import AbstractDataChunkIterator
from ...spec import RefSpec, DtypeSpec, NamespaceCatalog
from ...utils import docval, getargs, popargs, get_data_shape, get_docval, StrDataset
from ..utils import NamespaceToBuilderHelper, WriteStatusTracker

ROOT_NAME = 'root'
SPEC_LOC_ATTR = '.specloc'
H5_TEXT = special_dtype(vlen=str)
H5_BINARY = special_dtype(vlen=bytes)
H5_REF = special_dtype(ref=Reference)
H5_REGREF = special_dtype(ref=RegionReference)

H5PY_3 = h5py.__version__.startswith('3')


class HDF5IO(HDMFIO):

    __ns_spec_path = 'namespace'  # path to the namespace dataset within a namespace group

    @docval({'name': 'path', 'type': (str, Path), 'doc': 'the path to the HDF5 file', 'default': None},
            {'name': 'mode', 'type': str,
             'doc': ('the mode to open the HDF5 file with, one of ("w", "r", "r+", "a", "w-", "x"). '
                     'See `h5py.File <http://docs.h5py.org/en/latest/high/file.html#opening-creating-files>`_ for '
                     'more details.'),
             'default': 'r'},
            {'name': 'manager', 'type': (TypeMap, BuildManager),
             'doc': 'the BuildManager or a TypeMap to construct a BuildManager to use for I/O', 'default': None},
            {'name': 'comm', 'type': 'Intracomm',
             'doc': 'the MPI communicator to use for parallel I/O', 'default': None},
            {'name': 'file', 'type': [File, "S3File"], 'doc': 'a pre-existing h5py.File object', 'default': None},
            {'name': 'driver', 'type': str, 'doc': 'driver for h5py to use when opening HDF5 file', 'default': None})
    def __init__(self, **kwargs):
        """Open an HDF5 file for IO.
        """
        self.logger = logging.getLogger('%s.%s' % (self.__class__.__module__, self.__class__.__qualname__))
        path, manager, mode, comm, file_obj, driver = popargs('path', 'manager', 'mode', 'comm', 'file', 'driver',
                                                              kwargs)

        self.__open_links = []  # keep track of other files opened from links in this file
        self.__file = None  # This will be set below, but set to None first in case an error occurs and we need to close

        if path is None and file_obj is None:
            raise ValueError("You must supply either a path or a file.")

        if isinstance(path, Path):
            path = str(path)

        if file_obj is not None and path is not None and os.path.abspath(file_obj.filename) != os.path.abspath(path):
            msg = 'You argued %s as this object\'s path, ' % path
            msg += 'but supplied a file with filename: %s' % file_obj.filename
            raise ValueError(msg)

        if file_obj is None and not os.path.exists(path) and (mode == 'r' or mode == 'r+') and driver != 'ros3':
            msg = "Unable to open file %s in '%s' mode. File does not exist." % (path, mode)
            raise UnsupportedOperation(msg)

        if file_obj is None and os.path.exists(path) and (mode == 'w-' or mode == 'x'):
            msg = "Unable to open file %s in '%s' mode. File already exists." % (path, mode)
            raise UnsupportedOperation(msg)

        if manager is None:
            manager = BuildManager(TypeMap(NamespaceCatalog()))
        elif isinstance(manager, TypeMap):
            manager = BuildManager(manager)
        self.__driver = driver
        self.__comm = comm
        self.__mode = mode
        self.__file = file_obj
        super().__init__(manager, source=path)
        self.__built = dict()       # keep track of each builder for each dataset/group/link for each file
        self.__read = dict()        # keep track of which files have been read. Key is the filename value is the builder
        self.__ref_queue = deque()  # a queue of the references that need to be added
        self.__dci_queue = HDF5IODataChunkIteratorQueue()  # a queue of DataChunkIterators that need to be exhausted
        ObjectMapper.no_convert(Dataset)
        self._written_builders = WriteStatusTracker()  # track which builders were written (or read) by this IO object

    @property
    def comm(self):
        """The MPI communicator to use for parallel I/O."""
        return self.__comm

    @property
    def _file(self):
        return self.__file

    @property
    def driver(self):
        return self.__driver

    @staticmethod
    def __resolve_file_obj(path, file_obj, driver):
        if isinstance(path, Path):
            path = str(path)

        if path is None and file_obj is None:
            raise ValueError("Either the 'path' or 'file' argument must be supplied.")

        if path is not None and file_obj is not None:  # consistency check
            if os.path.abspath(file_obj.filename) != os.path.abspath(path):
                msg = ("You argued '%s' as this object's path, but supplied a file with filename: %s"
                       % (path, file_obj.filename))
                raise ValueError(msg)

        if file_obj is None:
            file_kwargs = dict()
            if driver is not None:
                file_kwargs.update(driver=driver)
            file_obj = File(path, 'r', **file_kwargs)
        return file_obj

    @classmethod
    @docval({'name': 'namespace_catalog', 'type': (NamespaceCatalog, TypeMap),
             'doc': 'the NamespaceCatalog or TypeMap to load namespaces into'},
            {'name': 'path', 'type': (str, Path), 'doc': 'the path to the HDF5 file', 'default': None},
            {'name': 'namespaces', 'type': list, 'doc': 'the namespaces to load', 'default': None},
            {'name': 'file', 'type': File, 'doc': 'a pre-existing h5py.File object', 'default': None},
            {'name': 'driver', 'type': str, 'doc': 'driver for h5py to use when opening HDF5 file', 'default': None},
            returns=("dict mapping the names of the loaded namespaces to a dict mapping included namespace names and "
                     "the included data types"),
            rtype=dict)
    def load_namespaces(cls, **kwargs):
        """Load cached namespaces from a file.

        If `file` is not supplied, then an :py:class:`h5py.File` object will be opened for the given `path`, the
        namespaces will be read, and the File object will be closed. If `file` is supplied, then
        the given File object will be read from and not closed.

        :raises ValueError: if both `path` and `file` are supplied but `path` is not the same as the path of `file`.
        """
        namespace_catalog, path, namespaces, file_obj, driver = popargs(
            'namespace_catalog', 'path', 'namespaces', 'file', 'driver', kwargs)

        open_file_obj = cls.__resolve_file_obj(path, file_obj, driver)
        if file_obj is None:  # need to close the file object that we just opened
            with open_file_obj:
                return cls.__load_namespaces(namespace_catalog, namespaces, open_file_obj)
        return cls.__load_namespaces(namespace_catalog, namespaces, open_file_obj)

    @classmethod
    def __load_namespaces(cls, namespace_catalog, namespaces, file_obj):
        d = {}

        if not cls.__check_specloc(file_obj):
            return d

        namespace_versions = cls.__get_namespaces(file_obj)

        spec_group = file_obj[file_obj.attrs[SPEC_LOC_ATTR]]
        if namespaces is None:
            namespaces = list(spec_group.keys())

        readers = dict()
        deps = dict()
        for ns in namespaces:
            latest_version = namespace_versions[ns]
            ns_group = spec_group[ns][latest_version]
            reader = H5SpecReader(ns_group)
            readers[ns] = reader
            # for each namespace in the 'namespace' dataset, track all included namespaces (dependencies)
            for spec_ns in reader.read_namespace(cls.__ns_spec_path):
                deps[ns] = list()
                for s in spec_ns['schema']:
                    dep = s.get('namespace')
                    if dep is not None:
                        deps[ns].append(dep)

        order = cls._order_deps(deps)
        for ns in order:
            reader = readers[ns]
            d.update(namespace_catalog.load_namespaces(cls.__ns_spec_path, reader=reader))

        return d

    @classmethod
    def __check_specloc(cls, file_obj):
        if SPEC_LOC_ATTR not in file_obj.attrs:
            # this occurs in legacy files
            msg = "No cached namespaces found in %s" % file_obj.filename
            warnings.warn(msg)
            return False
        return True

    @classmethod
    @docval({'name': 'path', 'type': (str, Path), 'doc': 'the path to the HDF5 file', 'default': None},
            {'name': 'file', 'type': File, 'doc': 'a pre-existing h5py.File object', 'default': None},
            {'name': 'driver', 'type': str, 'doc': 'driver for h5py to use when opening HDF5 file', 'default': None},
            returns="dict mapping names to versions of the namespaces in the file", rtype=dict)
    def get_namespaces(cls, **kwargs):
        """Get the names and versions of the cached namespaces from a file.

        If ``file`` is not supplied, then an :py:class:`h5py.File` object will be opened for the given ``path``, the
        namespaces will be read, and the File object will be closed. If `file` is supplied, then
        the given File object will be read from and not closed.

        If there are multiple versions of a namespace cached in the file, then only the latest one (using alphanumeric
        ordering) is returned. This is the version of the namespace that is loaded by HDF5IO.load_namespaces(...).

        :raises ValueError: if both `path` and `file` are supplied but `path` is not the same as the path of `file`.
        """
        path, file_obj, driver = popargs('path', 'file', 'driver', kwargs)

        open_file_obj = cls.__resolve_file_obj(path, file_obj, driver)
        if file_obj is None:  # need to close the file object that we just opened
            with open_file_obj:
                return cls.__get_namespaces(open_file_obj)
        return cls.__get_namespaces(open_file_obj)

    @classmethod
    def __get_namespaces(cls, file_obj):
        """Return a dict mapping namespace name to version string for the latest version of that namespace in the file.

        If there are multiple versions of a namespace cached in the file, then only the latest one (using alphanumeric
        ordering) is returned. This is the version of the namespace that is loaded by ``HDF5IO.load_namespaces``.
        """
        used_version_names = dict()
        if not cls.__check_specloc(file_obj):
            return used_version_names

        spec_group = file_obj[file_obj.attrs[SPEC_LOC_ATTR]]
        namespaces = list(spec_group.keys())
        for ns in namespaces:
            ns_group = spec_group[ns]
            # NOTE: by default, objects within groups are iterated in alphanumeric order
            version_names = list(ns_group.keys())
            if len(version_names) > 1:
                # prior to HDMF 1.6.1, extensions without a version were written under the group name "unversioned"
                # make sure that if there is another group representing a newer version, that is read instead
                if 'unversioned' in version_names:
                    version_names.remove('unversioned')
            if len(version_names) > 1:
                # as of HDMF 1.6.1, extensions without a version are written under the group name "None"
                # make sure that if there is another group representing a newer version, that is read instead
                if 'None' in version_names:
                    version_names.remove('None')
            used_version_names[ns] = version_names[-1]  # save the largest in alphanumeric order

        return used_version_names

    @classmethod
    def _order_deps(cls, deps):
        """
        Order namespaces according to dependency for loading into a NamespaceCatalog

        Args:
            deps (dict): a dictionary that maps a namespace name to a list of name of
                         the namespaces on which the namespace is directly dependent
                         Example: {'a': ['b', 'c'], 'b': ['d'], 'c': ['d'], 'd': []}
                         Expected output: ['d', 'b', 'c', 'a']
        """
        order = list()
        keys = list(deps.keys())
        deps = dict(deps)
        for k in keys:
            if k in deps:
                cls.__order_deps_aux(order, deps, k)
        return order

    @classmethod
    def __order_deps_aux(cls, order, deps, key):
        """
        A recursive helper function for _order_deps
        """
        if key not in deps:
            return
        subdeps = deps.pop(key)
        for subk in subdeps:
            cls.__order_deps_aux(order, deps, subk)
        order.append(key)

    @classmethod
    @docval({'name': 'source_filename', 'type': str, 'doc': 'the path to the HDF5 file to copy'},
            {'name': 'dest_filename', 'type': str, 'doc': 'the name of the destination file'},
            {'name': 'expand_external', 'type': bool, 'doc': 'expand external links into new objects', 'default': True},
            {'name': 'expand_refs', 'type': bool, 'doc': 'copy objects which are pointed to by reference',
             'default': False},
            {'name': 'expand_soft', 'type': bool, 'doc': 'expand soft links into new objects', 'default': False}
            )
    def copy_file(self, **kwargs):
        """
        Convenience function to copy an HDF5 file while allowing external links to be resolved.

        .. warning::

            As of HDMF 2.0, this method is no longer supported and may be removed in a future version.
            Please use the export method or h5py.File.copy method instead.

        .. note::

            The source file will be opened in 'r' mode and the destination file will be opened in 'w' mode
            using h5py. To avoid possible collisions, care should be taken that, e.g., the source file is
            not opened already when calling this function.

        """

        warnings.warn("The copy_file class method is no longer supported and may be removed in a future version of "
                      "HDMF. Please use the export method or h5py.File.copy method instead.", DeprecationWarning)

        source_filename, dest_filename, expand_external, expand_refs, expand_soft = getargs('source_filename',
                                                                                            'dest_filename',
                                                                                            'expand_external',
                                                                                            'expand_refs',
                                                                                            'expand_soft',
                                                                                            kwargs)
        source_file = File(source_filename, 'r')
        dest_file = File(dest_filename, 'w')
        for objname in source_file["/"].keys():
            source_file.copy(source=objname,
                             dest=dest_file,
                             name=objname,
                             expand_external=expand_external,
                             expand_refs=expand_refs,
                             expand_soft=expand_soft,
                             shallow=False,
                             without_attrs=False,
                             )
        for objname in source_file['/'].attrs:
            dest_file['/'].attrs[objname] = source_file['/'].attrs[objname]
        source_file.close()
        dest_file.close()

    @docval({'name': 'container', 'type': Container, 'doc': 'the Container object to write'},
            {'name': 'cache_spec', 'type': bool,
             'doc': ('If True (default), cache specification to file (highly recommended). If False, do not cache '
                     'specification to file. The appropriate specification will then need to be loaded prior to '
                     'reading the file.'),
             'default': True},
            {'name': 'link_data', 'type': bool,
             'doc': 'If True (default), create external links to HDF5 Datasets. If False, copy HDF5 Datasets.',
             'default': True},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'If True (default), exhaust DataChunkIterators one at a time. If False, exhaust them concurrently.',
             'default': True})
    def write(self, **kwargs):
        """Write the container to an HDF5 file."""
        if self.__mode == 'r':
            raise UnsupportedOperation(("Cannot write to file %s in mode '%s'. "
                                        "Please use mode 'r+', 'w', 'w-', 'x', or 'a'")
                                       % (self.source, self.__mode))

        cache_spec = popargs('cache_spec', kwargs)
        super().write(**kwargs)
        if cache_spec:
            self.__cache_spec()

    def __cache_spec(self):
        ref = self.__file.attrs.get(SPEC_LOC_ATTR)
        spec_group = None
        if ref is not None:
            spec_group = self.__file[ref]
        else:
            path = 'specifications'  # do something to figure out where the specifications should go
            spec_group = self.__file.require_group(path)
            self.__file.attrs[SPEC_LOC_ATTR] = spec_group.ref
        ns_catalog = self.manager.namespace_catalog
        for ns_name in ns_catalog.namespaces:
            ns_builder = NamespaceToBuilderHelper.convert_namespace(ns_catalog, ns_name)
            namespace = ns_catalog.get_namespace(ns_name)
            group_name = '%s/%s' % (ns_name, namespace.version)
            if group_name in spec_group:
                continue
            ns_group = spec_group.create_group(group_name)
            writer = H5SpecWriter(ns_group)
            ns_builder.export(self.__ns_spec_path, writer=writer)

    _export_args = (
        {'name': 'src_io', 'type': 'HDMFIO', 'doc': 'the HDMFIO object for reading the data to export'},
        {'name': 'container', 'type': Container,
         'doc': ('the Container object to export. If None, then the entire contents of the HDMFIO object will be '
                 'exported'),
         'default': None},
        {'name': 'write_args', 'type': dict, 'doc': 'arguments to pass to :py:meth:`write_builder`',
         'default': None},
        {'name': 'cache_spec', 'type': bool, 'doc': 'whether to cache the specification to file',
         'default': True}
    )

    @docval(*_export_args)
    def export(self, **kwargs):
        """Export data read from a file from any backend to HDF5.

        See :py:meth:`hdmf.backends.io.HDMFIO.export` for more details.
        """
        if self.__mode != 'w':
            raise UnsupportedOperation("Cannot export to file %s in mode '%s'. Please use mode 'w'."
                                       % (self.source, self.__mode))

        src_io = getargs('src_io', kwargs)
        write_args, cache_spec = popargs('write_args', 'cache_spec', kwargs)
        if write_args is None:
            write_args = dict()

        if not isinstance(src_io, HDF5IO) and write_args.get('link_data', True):
            raise UnsupportedOperation("Cannot export from non-HDF5 backend %s to HDF5 with write argument "
                                       "link_data=True." % src_io.__class__.__name__)

        write_args['export_source'] = src_io.source  # pass export_source=src_io.source to write_builder
        ckwargs = kwargs.copy()
        ckwargs['write_args'] = write_args
        super().export(**ckwargs)
        if cache_spec:
            self.__cache_spec()

    @classmethod
    @docval({'name': 'path', 'type': str, 'doc': 'the path to the destination HDF5 file'},
            {'name': 'comm', 'type': 'Intracomm', 'doc': 'the MPI communicator to use for parallel I/O',
             'default': None},
            *_export_args)  # NOTE: src_io is required and is the second positional argument
    def export_io(self, **kwargs):
        """Export from one backend to HDF5 (class method).

        Convenience function for :py:meth:`export` where you do not need to
        instantiate a new ``HDF5IO`` object for writing. An ``HDF5IO`` object is created with mode 'w' and the given
        arguments.

        Example usage:

        .. code-block:: python

            old_io = HDF5IO('old.h5', 'r')
            HDF5IO.export_io(path='new_copy.h5', src_io=old_io)

        See :py:meth:`export` for more details.
        """
        path, comm = popargs('path', 'comm', kwargs)

        with HDF5IO(path=path, comm=comm, mode='w') as write_io:
            write_io.export(**kwargs)

    def read(self, **kwargs):
        if self.__mode == 'w' or self.__mode == 'w-' or self.__mode == 'x':
            raise UnsupportedOperation("Cannot read from file %s in mode '%s'. Please use mode 'r', 'r+', or 'a'."
                                       % (self.source, self.__mode))
        try:
            return super().read(**kwargs)
        except UnsupportedOperation as e:
            if str(e) == 'Cannot build data. There are no values.':  # pragma: no cover
                raise UnsupportedOperation("Cannot read data from file %s in mode '%s'. There are no values."
                                           % (self.source, self.__mode))

    @docval(returns='a GroupBuilder representing the data object', rtype='GroupBuilder')
    def read_builder(self):
        """
        Read data and return the GroupBuilder representing it.

        NOTE: On read, the Builder.source may will usually not be set of the Builders.
        NOTE: The Builder.location is used internally to ensure correct handling of links (in particular on export)
        and should be set on read for all GroupBuilder, DatasetBuilder, and LinkBuilder objects.
        """
        if not self.__file:
            raise UnsupportedOperation("Cannot read data from closed HDF5 file '%s'" % self.source)
        f_builder = self.__read.get(self.__file)
        # ignore cached specs when reading builder
        ignore = set()
        specloc = self.__file.attrs.get(SPEC_LOC_ATTR)
        if specloc is not None:
            ignore.add(self.__file[specloc].name)
        if f_builder is None:
            f_builder = self.__read_group(self.__file, ROOT_NAME, ignore=ignore)
            self.__read[self.__file] = f_builder
        return f_builder

    def __set_written(self, builder):
        """
        Helper function used to set the written status for builders

        :param builder: Builder object to be marked as written
        :type builder: Builder
        """
        self._written_builders.set_written(builder)

    def get_written(self, builder):
        """Return True if this builder has been written to (or read from) disk by this IO object, False otherwise.

        :param builder: Builder object to get the written flag for
        :type builder: Builder

        :return: True if the builder is found in self._written_builders using the builder ID, False otherwise
        """
        return self._written_builders.get_written(builder)

    def __set_built(self, fpath, id, builder):
        """
        Update self.__built to cache the given builder for the given file and id.

        :param fpath: Path to the HDF5 file containing the object
        :type fpath: str
        :param id: ID of the HDF5 object in the path
        :type id: h5py GroupID object
        :param builder: The builder to be cached
        """
        self.__built.setdefault(fpath, dict()).setdefault(id, builder)

    def __get_built(self, fpath, id):
        """
        Look up a builder for the given file and id in self.__built cache

        :param fpath: Path to the HDF5 file containing the object
        :type fpath: str
        :param id: ID of the HDF5 object in the path
        :type id: h5py GroupID object

        :return: Builder in the self.__built cache or None
        """

        fdict = self.__built.get(fpath)
        if fdict:
            return fdict.get(id)
        else:
            return None

    @docval({'name': 'h5obj', 'type': (Dataset, Group),
             'doc': 'the HDF5 object to the corresponding Builder object for'})
    def get_builder(self, **kwargs):
        """
        Get the builder for the corresponding h5py Group or Dataset

        :raises ValueError: When no builder has been constructed yet for the given h5py object
        """
        h5obj = getargs('h5obj', kwargs)
        fpath = h5obj.file.filename
        builder = self.__get_built(fpath, h5obj.id)
        if builder is None:
            msg = '%s:%s has not been built' % (fpath, h5obj.name)
            raise ValueError(msg)
        return builder

    @docval({'name': 'h5obj', 'type': (Dataset, Group),
             'doc': 'the HDF5 object to the corresponding Container/Data object for'})
    def get_container(self, **kwargs):
        """
        Get the container for the corresponding h5py Group or Dataset

        :raises ValueError: When no builder has been constructed yet for the given h5py object
        """
        h5obj = getargs('h5obj', kwargs)
        builder = self.get_builder(h5obj)
        container = self.manager.construct(builder)
        return container

    def __read_group(self, h5obj, name=None, ignore=set()):
        kwargs = {
            "attributes": self.__read_attrs(h5obj),
            "groups": dict(),
            "datasets": dict(),
            "links": dict()
        }

        for key, val in kwargs['attributes'].items():
            if isinstance(val, bytes):
                kwargs['attributes'][key] = val.decode('UTF-8')

        if name is None:
            name = str(os.path.basename(h5obj.name))
        for k in h5obj:
            sub_h5obj = h5obj.get(k)
            if not (sub_h5obj is None):
                if sub_h5obj.name in ignore:
                    continue
                link_type = h5obj.get(k, getlink=True)
                if isinstance(link_type, SoftLink) or isinstance(link_type, ExternalLink):
                    # Reading links might be better suited in its own function
                    # get path of link (the key used for tracking what's been built)
                    target_path = link_type.path
                    target_obj = sub_h5obj.file[target_path]
                    builder_name = os.path.basename(target_path)
                    # get builder if already read, else build it
                    builder = self.__get_built(sub_h5obj.file.filename, target_obj.id)
                    if builder is None:
                        # NOTE: all links must have absolute paths
                        if isinstance(target_obj, Dataset):
                            builder = self.__read_dataset(target_obj, builder_name)
                        else:
                            builder = self.__read_group(target_obj, builder_name, ignore=ignore)
                        self.__set_built(sub_h5obj.file.filename,  target_obj.id, builder)
                    link_builder = LinkBuilder(builder=builder, name=k, source=h5obj.file.filename)
                    link_builder.location = h5obj.name
                    self.__set_written(link_builder)
                    kwargs['links'][builder_name] = link_builder
                    if isinstance(link_type, ExternalLink):
                        self.__open_links.append(sub_h5obj)
                else:
                    builder = self.__get_built(sub_h5obj.file.filename, sub_h5obj.id)
                    obj_type = None
                    read_method = None
                    if isinstance(sub_h5obj, Dataset):
                        read_method = self.__read_dataset
                        obj_type = kwargs['datasets']
                    else:
                        read_method = partial(self.__read_group, ignore=ignore)
                        obj_type = kwargs['groups']
                    if builder is None:
                        builder = read_method(sub_h5obj)
                        self.__set_built(sub_h5obj.file.filename, sub_h5obj.id, builder)
                    obj_type[builder.name] = builder
            else:
                warnings.warn(os.path.join(h5obj.name, k), BrokenLinkWarning)
                kwargs['datasets'][k] = None
                continue
        kwargs['source'] = h5obj.file.filename
        ret = GroupBuilder(name, **kwargs)
        ret.location = os.path.dirname(h5obj.name)
        self.__set_written(ret)
        return ret

    def __read_dataset(self, h5obj, name=None):
        kwargs = {
            "attributes": self.__read_attrs(h5obj),
            "dtype": h5obj.dtype,
            "maxshape": h5obj.maxshape
        }
        for key, val in kwargs['attributes'].items():
            if isinstance(val, bytes):
                kwargs['attributes'][key] = val.decode('UTF-8')

        if name is None:
            name = str(os.path.basename(h5obj.name))
        kwargs['source'] = h5obj.file.filename
        ndims = len(h5obj.shape)
        if ndims == 0:                                       # read scalar
            scalar = h5obj[()]
            if isinstance(scalar, bytes):
                scalar = scalar.decode('UTF-8')

            if isinstance(scalar, Reference):
                # TODO (AJTRITT):  This should call __read_ref to support Group references
                target = h5obj.file[scalar]
                target_builder = self.__read_dataset(target)
                self.__set_built(target.file.filename, target.id, target_builder)
                if isinstance(scalar, RegionReference):
                    d = RegionBuilder(scalar, target_builder)
                else:
                    d = ReferenceBuilder(target_builder)
                kwargs['data'] = d
                kwargs['dtype'] = d.dtype
            else:
                kwargs["data"] = scalar
        else:
            d = None
            if h5obj.dtype.kind == 'O' and len(h5obj) > 0:
                elem1 = h5obj[tuple([0] * (h5obj.ndim - 1) + [0])]
                if isinstance(elem1, (str, bytes)):
                    d = self._check_str_dtype(h5obj)
                elif isinstance(elem1, RegionReference):  # read list of references
                    d = BuilderH5RegionDataset(h5obj, self)
                    kwargs['dtype'] = d.dtype
                elif isinstance(elem1, Reference):
                    d = BuilderH5ReferenceDataset(h5obj, self)
                    kwargs['dtype'] = d.dtype
            elif h5obj.dtype.kind == 'V':    # table / compound data type
                cpd_dt = h5obj.dtype
                ref_cols = [check_dtype(ref=cpd_dt[i]) or check_dtype(vlen=cpd_dt[i]) for i in range(len(cpd_dt))]
                d = BuilderH5TableDataset(h5obj, self, ref_cols)
                kwargs['dtype'] = HDF5IO.__compound_dtype_to_list(h5obj.dtype, d.dtype)
            else:
                d = h5obj
            kwargs["data"] = d
        ret = DatasetBuilder(name, **kwargs)
        ret.location = os.path.dirname(h5obj.name)
        self.__set_written(ret)
        return ret

    def _check_str_dtype(self, h5obj):
        dtype = h5obj.dtype
        if dtype.kind == 'O':
            if dtype.metadata.get('vlen') == str and H5PY_3:
                return StrDataset(h5obj, None)
        return h5obj

    @classmethod
    def __compound_dtype_to_list(cls, h5obj_dtype, dset_dtype):
        ret = []
        for name, dtype in zip(h5obj_dtype.fields, dset_dtype):
            ret.append({'name': name, 'dtype': dtype})
        return ret

    def __read_attrs(self, h5obj):
        ret = dict()
        for k, v in h5obj.attrs.items():
            if k == SPEC_LOC_ATTR:     # ignore cached spec
                continue
            if isinstance(v, RegionReference):
                raise ValueError("cannot read region reference attributes yet")
            elif isinstance(v, Reference):
                ret[k] = self.__read_ref(h5obj.file[v])
            else:
                ret[k] = v
        return ret

    def __read_ref(self, h5obj):
        ret = None
        ret = self.__get_built(h5obj.file.filename, h5obj.id)
        if ret is None:
            if isinstance(h5obj, Dataset):
                ret = self.__read_dataset(h5obj)
            elif isinstance(h5obj, Group):
                ret = self.__read_group(h5obj)
            else:
                raise ValueError("h5obj must be a Dataset or a Group - got %s" % str(h5obj))
            self.__set_built(h5obj.file.filename, h5obj.id, ret)
        return ret

    def open(self):
        if self.__file is None:
            open_flag = self.__mode
            kwargs = dict()
            if self.comm:
                kwargs.update(driver='mpio', comm=self.comm)

            if self.driver is not None:
                kwargs.update(driver=self.driver)

            self.__file = File(self.source, open_flag, **kwargs)

    def close(self, close_links=True):
        """Close this file and any files linked to from this file.

        :param close_links: Whether to close all files linked to from this file. (default: True)
        :type close_links: bool
        """
        if close_links:
            self.close_linked_files()
        try:
            if self.__file is not None:
                self.__file.close()
        except AttributeError:
            # Do not do anything in case that self._file does not exist. This
            # may happen in case that an error occurs before HDF5IO has been fully
            # setup in __init__, e.g,. if a child class (such as NWBHDF5IO) raises
            # an error before self.__file has been created
            self.__file = None

    def close_linked_files(self):
        """Close all opened, linked-to files.

        MacOS and Linux automatically release the linked-to file after the linking file is closed, but Windows does
        not, which prevents the linked-to file from being deleted or truncated. Use this method to close all opened,
        linked-to files.
        """
        # Make sure
        try:
            for obj in self.__open_links:
                if obj:
                    obj.file.close()
        except AttributeError:
            # Do not do anything in case that self.__open_links does not exist. This
            # may happen in case that an error occurs before HDF5IO has been fully
            # setup in __init__, e.g,. if a child class (such as NWBHDF5IO) raises
            # an error before self.__open_links has been created.
            pass
        finally:
            self.__open_links = []

    @docval({'name': 'builder', 'type': GroupBuilder, 'doc': 'the GroupBuilder object representing the HDF5 file'},
            {'name': 'link_data', 'type': bool,
             'doc': 'If not specified otherwise link (True) or copy (False) HDF5 Datasets', 'default': True},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'exhaust DataChunkIterators one at a time. If False, exhaust them concurrently',
             'default': True},
            {'name': 'export_source', 'type': str,
             'doc': 'The source of the builders when exporting', 'default': None})
    def write_builder(self, **kwargs):
        f_builder = popargs('builder', kwargs)
        link_data, exhaust_dci, export_source = getargs('link_data', 'exhaust_dci', 'export_source', kwargs)
        self.logger.debug("Writing GroupBuilder '%s' to path '%s' with kwargs=%s"
                          % (f_builder.name, self.source, kwargs))
        for name, gbldr in f_builder.groups.items():
            self.write_group(self.__file, gbldr, **kwargs)
        for name, dbldr in f_builder.datasets.items():
            self.write_dataset(self.__file, dbldr, **kwargs)
        for name, lbldr in f_builder.links.items():
            self.write_link(self.__file, lbldr, export_source=kwargs.get("export_source"))
        self.set_attributes(self.__file, f_builder.attributes)
        self.__add_refs()
        self.__dci_queue.exhaust_queue()
        self.__set_written(f_builder)
        self.logger.debug("Done writing %s '%s' to path '%s'" %
                          (f_builder.__class__.__qualname__, f_builder.name, self.source))

    def __add_refs(self):
        '''
        Add all references in the file.

        References get queued to be added at the end of write. This is because
        the current traversal algorithm (i.e. iterating over GroupBuilder items)
        does not happen in a guaranteed order. We need to figure out what objects
        will be references, and then write them after we write everything else.
        '''
        failed = set()
        while len(self.__ref_queue) > 0:
            call = self.__ref_queue.popleft()
            self.logger.debug("Adding reference with call id %d from queue (length %d)"
                              % (id(call), len(self.__ref_queue)))
            try:
                call()
            except KeyError:
                if id(call) in failed:
                    raise RuntimeError('Unable to resolve reference')
                self.logger.debug("Adding reference with call id %d failed. Appending call to queue" % id(call))
                failed.add(id(call))
                self.__ref_queue.append(call)

    @classmethod
    def get_type(cls, data):
        if isinstance(data, str):
            return H5_TEXT
        elif isinstance(data, bytes):
            return H5_BINARY
        elif isinstance(data, Container):
            return H5_REF
        elif not hasattr(data, '__len__'):
            return type(data)
        else:
            if len(data) == 0:
                if hasattr(data, 'dtype'):
                    return data.dtype
                else:
                    raise ValueError('cannot determine type for empty data')
            return cls.get_type(data[0])

    __dtypes = {
        "float": np.float32,
        "float32": np.float32,
        "double": np.float64,
        "float64": np.float64,
        "long": np.int64,
        "int64": np.int64,
        "int": np.int32,
        "int32": np.int32,
        "short": np.int16,
        "int16": np.int16,
        "int8": np.int8,
        "uint64": np.uint64,
        "uint": np.uint32,
        "uint32": np.uint32,
        "uint16": np.uint16,
        "uint8": np.uint8,
        "bool": np.bool_,
        "text": H5_TEXT,
        "utf": H5_TEXT,
        "utf8": H5_TEXT,
        "utf-8": H5_TEXT,
        "ascii": H5_BINARY,
        "bytes": H5_BINARY,
        "ref": H5_REF,
        "reference": H5_REF,
        "object": H5_REF,
        "region": H5_REGREF,
        "isodatetime": H5_TEXT,
        "datetime": H5_TEXT,
    }

    @classmethod
    def __resolve_dtype__(cls, dtype, data):
        # TODO: These values exist, but I haven't solved them yet
        # binary
        # number
        dtype = cls.__resolve_dtype_helper__(dtype)
        if dtype is None:
            dtype = cls.get_type(data)
        return dtype

    @classmethod
    def __resolve_dtype_helper__(cls, dtype):
        if dtype is None:
            return None
        elif isinstance(dtype, str):
            return cls.__dtypes.get(dtype)
        elif isinstance(dtype, dict):
            return cls.__dtypes.get(dtype['reftype'])
        elif isinstance(dtype, np.dtype):
            # NOTE: some dtypes may not be supported, but we need to support writing of read-in compound types
            return dtype
        else:
            return np.dtype([(x['name'], cls.__resolve_dtype_helper__(x['dtype'])) for x in dtype])

    @docval({'name': 'obj', 'type': (Group, Dataset), 'doc': 'the HDF5 object to add attributes to'},
            {'name': 'attributes',
             'type': dict,
             'doc': 'a dict containing the attributes on the Group or Dataset, indexed by attribute name'})
    def set_attributes(self, **kwargs):
        obj, attributes = getargs('obj', 'attributes', kwargs)
        for key, value in attributes.items():
            try:
                if isinstance(value, (set, list, tuple)):
                    tmp = tuple(value)
                    if len(tmp) > 0:
                        if isinstance(tmp[0], (str, bytes)):
                            value = np.array(value, dtype=special_dtype(vlen=type(tmp[0])))
                        elif isinstance(tmp[0], Container):  # a list of references
                            self.__queue_ref(self._make_attr_ref_filler(obj, key, tmp))
                        else:
                            value = np.array(value)
                    self.logger.debug("Setting %s '%s' attribute '%s' to %s"
                                      % (obj.__class__.__name__, obj.name, key, value.__class__.__name__))
                    obj.attrs[key] = value
                elif isinstance(value, (Container, Builder, ReferenceBuilder)):           # a reference
                    self.__queue_ref(self._make_attr_ref_filler(obj, key, value))
                else:
                    self.logger.debug("Setting %s '%s' attribute '%s' to %s"
                                      % (obj.__class__.__name__, obj.name, key, value.__class__.__name__))
                    if isinstance(value, np.ndarray) and value.dtype.kind == 'U':
                        value = np.array(value, dtype=H5_TEXT)
                    obj.attrs[key] = value                   # a regular scalar
            except Exception as e:
                msg = "unable to write attribute '%s' on object '%s'" % (key, obj.name)
                raise RuntimeError(msg) from e

    def _make_attr_ref_filler(self, obj, key, value):
        '''
            Make the callable for setting references to attributes
        '''
        self.logger.debug("Queueing set %s '%s' attribute '%s' to %s"
                          % (obj.__class__.__name__, obj.name, key, value.__class__.__name__))
        if isinstance(value, (tuple, list)):
            def _filler():
                ret = list()
                for item in value:
                    ret.append(self.__get_ref(item))
                obj.attrs[key] = ret
        else:
            def _filler():
                obj.attrs[key] = self.__get_ref(value)
        return _filler

    @docval({'name': 'parent', 'type': Group, 'doc': 'the parent HDF5 object'},
            {'name': 'builder', 'type': GroupBuilder, 'doc': 'the GroupBuilder to write'},
            {'name': 'link_data', 'type': bool,
             'doc': 'If not specified otherwise link (True) or copy (False) HDF5 Datasets', 'default': True},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'exhaust DataChunkIterators one at a time. If False, exhaust them concurrently',
             'default': True},
            {'name': 'export_source', 'type': str,
             'doc': 'The source of the builders when exporting', 'default': None},
            returns='the Group that was created', rtype='Group')
    def write_group(self, **kwargs):
        parent, builder = popargs('parent', 'builder', kwargs)
        self.logger.debug("Writing GroupBuilder '%s' to parent group '%s'" % (builder.name, parent.name))
        if self.get_written(builder):
            self.logger.debug("    GroupBuilder '%s' is already written" % builder.name)
            group = parent[builder.name]
        else:
            self.logger.debug("    Creating group '%s'" % builder.name)
            group = parent.create_group(builder.name)
        # write all groups
        subgroups = builder.groups
        if subgroups:
            for subgroup_name, sub_builder in subgroups.items():
                # do not create an empty group without attributes or links
                self.write_group(group, sub_builder, **kwargs)
        # write all datasets
        datasets = builder.datasets
        if datasets:
            for dset_name, sub_builder in datasets.items():
                self.write_dataset(group, sub_builder, **kwargs)
        # write all links
        links = builder.links
        if links:
            for link_name, sub_builder in links.items():
                self.write_link(group, sub_builder, export_source=kwargs.get("export_source"))
        attributes = builder.attributes
        self.set_attributes(group, attributes)
        self.__set_written(builder)
        return group

    def __get_path(self, builder):
        """Get the path to the builder.

        Note that the root of the file has no name - it is just "/". Thus, the name of the root container is ignored.
        If builder.location is set then it is used as the path, otherwise the function
        determines the path by constructing it iteratively from the parents of the
        builder.
        """
        if builder.location is not None:
            path = os.path.normpath(os.path.join(builder.location, builder.name)).replace("\\", "/")
        else:
            curr = builder
            names = list()
            while curr.parent is not None:
                names.append(curr.name)
                curr = curr.parent
            delim = "/"
            path = "%s%s" % (delim, delim.join(reversed(names)))
        return path

    @docval({'name': 'parent', 'type': Group, 'doc': 'the parent HDF5 object'},
            {'name': 'builder', 'type': LinkBuilder, 'doc': 'the LinkBuilder to write'},
            {'name': 'export_source', 'type': str,
             'doc': 'The source of the builders when exporting', 'default': None},
            returns='the Link that was created', rtype='Link')
    def write_link(self, **kwargs):
        parent, builder, export_source = getargs('parent', 'builder', 'export_source', kwargs)
        self.logger.debug("Writing LinkBuilder '%s' to parent group '%s'" % (builder.name, parent.name))
        if self.get_written(builder):
            self.logger.debug("    LinkBuilder '%s' is already written" % builder.name)
            return None
        name = builder.name
        target_builder = builder.builder
        path = self.__get_path(target_builder)
        # source will indicate target_builder's location
        if export_source is None:
            write_source = builder.source
        else:
            write_source = export_source

        if write_source == target_builder.source:
            link_obj = SoftLink(path)
            self.logger.debug("    Creating SoftLink '%s/%s' to '%s'"
                              % (parent.name, name, link_obj.path))
        elif target_builder.source is not None:
            target_filename = os.path.abspath(target_builder.source)
            parent_filename = os.path.abspath(parent.file.filename)
            relative_path = os.path.relpath(target_filename, os.path.dirname(parent_filename))
            if target_builder.location is not None:
                path = target_builder.location + "/" + target_builder.name
            link_obj = ExternalLink(relative_path, path)
            self.logger.debug("    Creating ExternalLink '%s/%s' to '%s://%s'"
                              % (parent.name, name, link_obj.filename, link_obj.path))
        else:
            msg = 'cannot create external link to %s' % path
            raise ValueError(msg)
        parent[name] = link_obj
        self.__set_written(builder)
        return link_obj

    @docval({'name': 'parent', 'type': Group, 'doc': 'the parent HDF5 object'},  # noqa: C901
            {'name': 'builder', 'type': DatasetBuilder, 'doc': 'the DatasetBuilder to write'},
            {'name': 'link_data', 'type': bool,
             'doc': 'If not specified otherwise link (True) or copy (False) HDF5 Datasets', 'default': True},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'exhaust DataChunkIterators one at a time. If False, exhaust them concurrently',
             'default': True},
            {'name': 'export_source', 'type': str,
             'doc': 'The source of the builders when exporting', 'default': None},
            returns='the Dataset that was created', rtype=Dataset)
    def write_dataset(self, **kwargs):  # noqa: C901
        """ Write a dataset to HDF5

        The function uses other dataset-dependent write functions, e.g,
        ``__scalar_fill__``, ``__list_fill__``, and ``__setup_chunked_dset__`` to write the data.
        """
        parent, builder = popargs('parent', 'builder', kwargs)
        link_data, exhaust_dci, export_source = getargs('link_data', 'exhaust_dci', 'export_source', kwargs)
        self.logger.debug("Writing DatasetBuilder '%s' to parent group '%s'" % (builder.name, parent.name))
        if self.get_written(builder):
            self.logger.debug("    DatasetBuilder '%s' is already written" % builder.name)
            return None
        name = builder.name
        data = builder.data
        dataio = None
        options = dict()   # dict with additional
        if isinstance(data, H5DataIO):
            options['io_settings'] = data.io_settings
            dataio = data
            link_data = data.link_data
            data = data.data
        else:
            options['io_settings'] = {}
        attributes = builder.attributes
        options['dtype'] = builder.dtype
        dset = None
        link = None

        # The user provided an existing h5py dataset as input and asked to create a link to the dataset
        if isinstance(data, Dataset):
            data_filename = os.path.abspath(data.file.filename)
            if link_data:
                if export_source is None:  # not exporting
                    parent_filename = os.path.abspath(parent.file.filename)
                    if data_filename != parent_filename:  # create external link to data
                        relative_path = os.path.relpath(data_filename, os.path.dirname(parent_filename))
                        link = ExternalLink(relative_path, data.name)
                        self.logger.debug("    Creating ExternalLink '%s/%s' to '%s://%s'"
                                          % (parent.name, name, link.filename, link.path))
                    else:  # create soft link to dataset already in this file -- possible if mode == 'r+'
                        link = SoftLink(data.name)
                        self.logger.debug("    Creating SoftLink '%s/%s' to '%s'"
                                          % (parent.name, name, link.path))
                    parent[name] = link
                else:  # exporting
                    export_source = os.path.abspath(export_source)
                    parent_filename = os.path.abspath(parent.file.filename)
                    if data_filename != export_source:  # dataset is in different file than export source
                        # possible if user adds a link to a dataset in a different file after reading export source
                        # to memory
                        relative_path = os.path.relpath(data_filename, os.path.dirname(parent_filename))
                        link = ExternalLink(relative_path, data.name)
                        self.logger.debug("    Creating ExternalLink '%s/%s' to '%s://%s'"
                                          % (parent.name, name, link.filename, link.path))
                        parent[name] = link
                    elif parent.name != data.parent.name:  # dataset is in export source and has different path
                        # so create a soft link to the dataset in this file
                        # possible if user adds a link to a dataset in export source after reading to memory
                        # TODO check that there is/will be still a dataset at data.name -- if the dataset has
                        # been removed, then this link will be broken
                        link = SoftLink(data.name)
                        self.logger.debug("    Creating SoftLink '%s/%s' to '%s'"
                                          % (parent.name, name, link.path))
                        parent[name] = link
                    else:  # dataset is in export source and has same path as the builder, so copy the dataset
                        self.logger.debug("    Copying data from '%s://%s' to '%s/%s'"
                                          % (data.file.filename, data.name, parent.name, name))
                        parent.copy(source=data,
                                    dest=parent,
                                    name=name,
                                    expand_soft=False,
                                    expand_external=False,
                                    expand_refs=False,
                                    without_attrs=True)
                        dset = parent[name]
            else:
                # TODO add option for case where there are multiple links to the same dataset within a file:
                # instead of copying the dset N times, copy it once and create soft links to it within the file
                self.logger.debug("    Copying data from '%s://%s' to '%s/%s'"
                                  % (data.file.filename, data.name, parent.name, name))
                parent.copy(source=data,
                            dest=parent,
                            name=name,
                            expand_soft=False,
                            expand_external=False,
                            expand_refs=False,
                            without_attrs=True)
                dset = parent[name]

        #  Write a compound dataset, i.e, a dataset with compound data type
        elif isinstance(options['dtype'], list):
            # do some stuff to figure out what data is a reference
            refs = list()
            for i, dts in enumerate(options['dtype']):
                if self.__is_ref(dts):
                    refs.append(i)
            # If one ore more of the parts of the compound data type are references then we need to deal with those
            if len(refs) > 0:
                try:
                    _dtype = self.__resolve_dtype__(options['dtype'], data)
                except Exception as exc:
                    msg = 'cannot add %s to %s - could not determine type' % (name, parent.name)
                    raise Exception(msg) from exc
                dset = parent.require_dataset(name, shape=(len(data),), dtype=_dtype, **options['io_settings'])
                self.__set_written(builder)
                self.logger.debug("Queueing reference resolution and set attribute on dataset '%s' containing "
                                  "object references. attributes: %s"
                                  % (name, list(attributes.keys())))

                @self.__queue_ref
                def _filler():
                    self.logger.debug("Resolving object references and setting attribute on dataset '%s' "
                                      "containing attributes: %s"
                                      % (name, list(attributes.keys())))
                    ret = list()
                    for item in data:
                        new_item = list(item)
                        for i in refs:
                            new_item[i] = self.__get_ref(item[i])
                        ret.append(tuple(new_item))
                    dset = parent[name]
                    dset[:] = ret
                    self.set_attributes(dset, attributes)

                return
            # If the compound data type contains only regular data (i.e., no references) then we can write it as usual
            else:
                dset = self.__list_fill__(parent, name, data, options)
        # Write a dataset containing references, i.e., a region or object reference.
        # NOTE: we can ignore options['io_settings'] for scalar data
        elif self.__is_ref(options['dtype']):
            _dtype = self.__dtypes.get(options['dtype'])
            # Write a scalar data region reference dataset
            if isinstance(data, RegionBuilder):
                dset = parent.require_dataset(name, shape=(), dtype=_dtype)
                self.__set_written(builder)
                self.logger.debug("Queueing reference resolution and set attribute on dataset '%s' containing a "
                                  "region reference. attributes: %s"
                                  % (name, list(attributes.keys())))

                @self.__queue_ref
                def _filler():
                    self.logger.debug("Resolving region reference and setting attribute on dataset '%s' "
                                      "containing attributes: %s"
                                      % (name, list(attributes.keys())))
                    ref = self.__get_ref(data.builder, data.region)
                    dset = parent[name]
                    dset[()] = ref
                    self.set_attributes(dset, attributes)
            # Write a scalar object reference dataset
            elif isinstance(data, ReferenceBuilder):
                dset = parent.require_dataset(name, dtype=_dtype, shape=())
                self.__set_written(builder)
                self.logger.debug("Queueing reference resolution and set attribute on dataset '%s' containing an "
                                  "object reference. attributes: %s"
                                  % (name, list(attributes.keys())))

                @self.__queue_ref
                def _filler():
                    self.logger.debug("Resolving object reference and setting attribute on dataset '%s' "
                                      "containing attributes: %s"
                                      % (name, list(attributes.keys())))
                    ref = self.__get_ref(data.builder)
                    dset = parent[name]
                    dset[()] = ref
                    self.set_attributes(dset, attributes)
            # Write an array dataset of references
            else:
                # Write a array of region references
                if options['dtype'] == 'region':
                    dset = parent.require_dataset(name, dtype=_dtype, shape=(len(data),), **options['io_settings'])
                    self.__set_written(builder)
                    self.logger.debug("Queueing reference resolution and set attribute on dataset '%s' containing "
                                      "region references. attributes: %s"
                                      % (name, list(attributes.keys())))

                    @self.__queue_ref
                    def _filler():
                        self.logger.debug("Resolving region references and setting attribute on dataset '%s' "
                                          "containing attributes: %s"
                                          % (name, list(attributes.keys())))
                        refs = list()
                        for item in data:
                            refs.append(self.__get_ref(item.builder, item.region))
                        dset = parent[name]
                        dset[()] = refs
                        self.set_attributes(dset, attributes)
                # Write array of object references
                else:
                    dset = parent.require_dataset(name, shape=(len(data),), dtype=_dtype, **options['io_settings'])
                    self.__set_written(builder)
                    self.logger.debug("Queueing reference resolution and set attribute on dataset '%s' containing "
                                      "object references. attributes: %s"
                                      % (name, list(attributes.keys())))

                    @self.__queue_ref
                    def _filler():
                        self.logger.debug("Resolving object references and setting attribute on dataset '%s' "
                                          "containing attributes: %s"
                                          % (name, list(attributes.keys())))
                        refs = list()
                        for item in data:
                            refs.append(self.__get_ref(item))
                        dset = parent[name]
                        dset[()] = refs
                        self.set_attributes(dset, attributes)
            return
        # write a "regular" dataset
        else:
            # Create an empty dataset
            if data is None:
                dset = self.__setup_empty_dset__(parent, name, options['io_settings'])
                dataio.dataset = dset
            # Write a scalar dataset containing a single string
            elif isinstance(data, (str, bytes)):
                dset = self.__scalar_fill__(parent, name, data, options)
            # Iterative write of a data chunk iterator
            elif isinstance(data, AbstractDataChunkIterator):
                dset = self.__setup_chunked_dset__(parent, name, data, options)
                self.__dci_queue.append(dataset=dset, data=data)
            # Write a regular in memory array (e.g., numpy array, list etc.)
            elif hasattr(data, '__len__'):
                dset = self.__list_fill__(parent, name, data, options)
            # Write a regular scalar dataset
            else:
                dset = self.__scalar_fill__(parent, name, data, options)
        # Create the attributes on the dataset only if we are the primary and not just a Soft/External link
        if link is None:
            self.set_attributes(dset, attributes)
        # Validate the attributes on the linked dataset
        elif len(attributes) > 0:
            pass
        self.__set_written(builder)
        if exhaust_dci:
            self.__dci_queue.exhaust_queue()

    @classmethod
    def __scalar_fill__(cls, parent, name, data, options=None):
        dtype = None
        io_settings = {}
        if options is not None:
            dtype = options.get('dtype')
            io_settings = options.get('io_settings')
        if not isinstance(dtype, type):
            try:
                dtype = cls.__resolve_dtype__(dtype, data)
            except Exception as exc:
                msg = 'cannot add %s to %s - could not determine type' % (name, parent.name)
                raise Exception(msg) from exc
        try:
            dset = parent.create_dataset(name, data=data, shape=None, dtype=dtype, **io_settings)
        except Exception as exc:
            msg = "Could not create scalar dataset %s in %s" % (name, parent.name)
            raise Exception(msg) from exc
        return dset

    @classmethod
    def __setup_chunked_dset__(cls, parent, name, data, options=None):
        """
        Setup a dataset for writing to one-chunk-at-a-time based on the given DataChunkIterator

        :param parent: The parent object to which the dataset should be added
        :type parent: h5py.Group, h5py.File
        :param name: The name of the dataset
        :type name: str
        :param data: The data to be written.
        :type data: DataChunkIterator
        :param options: Dict with options for creating a dataset. available options are 'dtype' and 'io_settings'
        :type options: dict

        """
        io_settings = {}
        if options is not None:
            if 'io_settings' in options:
                io_settings = options.get('io_settings')
        # Define the chunking options if the user has not set them explicitly. We need chunking for the iterative write.
        if 'chunks' not in io_settings:
            recommended_chunks = data.recommended_chunk_shape()
            io_settings['chunks'] = True if recommended_chunks is None else recommended_chunks
        # Define the shape of the data if not provided by the user
        if 'shape' not in io_settings:
            io_settings['shape'] = data.recommended_data_shape()
        # Define the maxshape of the data if not provided by the user
        if 'maxshape' not in io_settings:
            io_settings['maxshape'] = data.maxshape
        if 'dtype' not in io_settings:
            if (options is not None) and ('dtype' in options):
                io_settings['dtype'] = options['dtype']
            else:
                io_settings['dtype'] = data.dtype
            if isinstance(io_settings['dtype'], str):
                # map to real dtype if we were given a string
                io_settings['dtype'] = cls.__dtypes.get(io_settings['dtype'])
        try:
            dset = parent.create_dataset(name, **io_settings)
        except Exception as exc:
            raise Exception("Could not create dataset %s in %s" % (name, parent.name)) from exc
        return dset

    @classmethod
    def __setup_empty_dset__(cls, parent, name, io_settings):
        """
        Setup a dataset for writing to one-chunk-at-a-time based on the given DataChunkIterator

        :param parent: The parent object to which the dataset should be added
        :type parent: h5py.Group, h5py.File
        :param name: The name of the dataset
        :type name: str
        :param data: The data to be written.
        :type data: DataChunkIterator
        :param options: Dict with options for creating a dataset. available options are 'dtype' and 'io_settings'
        :type options: dict

        """
        # Define the shape of the data if not provided by the user
        if 'shape' not in io_settings:
            raise ValueError(f"Cannot setup empty dataset {pp(parent.name, name)} without shape")
        if 'dtype' not in io_settings:
            raise ValueError(f"Cannot setup empty dataset {pp(parent.name, name)} without dtype")
        if isinstance(io_settings['dtype'], str):
            # map to real dtype if we were given a string
            io_settings['dtype'] = cls.__dtypes.get(io_settings['dtype'])
        try:
            dset = parent.create_dataset(name, **io_settings)
        except Exception as exc:
            raise Exception("Could not create dataset %s in %s" % (name, parent.name)) from exc
        return dset

    @classmethod
    def __chunked_iter_fill__(cls, parent, name, data, options=None):
        """
        Write data to a dataset one-chunk-at-a-time based on the given DataChunkIterator

        :param parent: The parent object to which the dataset should be added
        :type parent: h5py.Group, h5py.File
        :param name: The name of the dataset
        :type name: str
        :param data: The data to be written.
        :type data: DataChunkIterator
        :param options: Dict with options for creating a dataset. available options are 'dtype' and 'io_settings'
        :type options: dict

        """
        dset = cls.__setup_chunked_dset__(parent, name, data, options=options)
        read = True
        while read:
            read = HDF5IODataChunkIteratorQueue._write_chunk(dset, data)
        return dset

    @classmethod
    def __list_fill__(cls, parent, name, data, options=None):
        # define the io settings and data type if necessary
        io_settings = {}
        dtype = None
        if options is not None:
            dtype = options.get('dtype')
            io_settings = options.get('io_settings')
        if not isinstance(dtype, type):
            try:
                dtype = cls.__resolve_dtype__(dtype, data)
            except Exception as exc:
                msg = 'cannot add %s to %s - could not determine type' % (name, parent.name)
                raise Exception(msg) from exc
        # define the data shape
        if 'shape' in io_settings:
            data_shape = io_settings.pop('shape')
        elif hasattr(data, 'shape'):
            data_shape = data.shape
        elif isinstance(dtype, np.dtype):
            data_shape = (len(data),)
        else:
            data_shape = get_data_shape(data)

        # Create the dataset
        try:
            dset = parent.create_dataset(name, shape=data_shape, dtype=dtype, **io_settings)
        except Exception as exc:
            msg = "Could not create dataset %s in %s with shape %s, dtype %s, and iosettings %s. %s" % \
                  (name, parent.name, str(data_shape), str(dtype), str(io_settings), str(exc))
            raise Exception(msg) from exc
        # Write the data
        if len(data) > dset.shape[0]:
            new_shape = list(dset.shape)
            new_shape[0] = len(data)
            dset.resize(new_shape)
        try:
            dset[:] = data
        except Exception as e:
            raise e
        return dset

    @docval({'name': 'container', 'type': (Builder, Container, ReferenceBuilder), 'doc': 'the object to reference',
             'default': None},
            {'name': 'region', 'type': (slice, list, tuple), 'doc': 'the region reference indexing object',
             'default': None},
            returns='the reference', rtype=Reference)
    def __get_ref(self, **kwargs):
        container, region = getargs('container', 'region', kwargs)
        if container is None:
            return None
        if isinstance(container, Builder):
            self.logger.debug("Getting reference for %s '%s'" % (container.__class__.__name__, container.name))
            if isinstance(container, LinkBuilder):
                builder = container.target_builder
            else:
                builder = container
        elif isinstance(container, ReferenceBuilder):
            self.logger.debug("Getting reference for %s '%s'" % (container.__class__.__name__, container.builder.name))
            builder = container.builder
        else:
            self.logger.debug("Getting reference for %s '%s'" % (container.__class__.__name__, container.name))
            builder = self.manager.build(container)
        path = self.__get_path(builder)
        self.logger.debug("Getting reference at path '%s'" % path)
        if isinstance(container, RegionBuilder):
            region = container.region
        if region is not None:
            dset = self.__file[path]
            if not isinstance(dset, Dataset):
                raise ValueError('cannot create region reference without Dataset')
            return self.__file[path].regionref[region]
        else:
            return self.__file[path].ref

    def __is_ref(self, dtype):
        if isinstance(dtype, DtypeSpec):
            return self.__is_ref(dtype.dtype)
        if isinstance(dtype, RefSpec):
            return True
        if isinstance(dtype, dict):  # may be dict from reading a compound dataset
            return self.__is_ref(dtype['dtype'])
        if isinstance(dtype, str):
            return dtype == DatasetBuilder.OBJECT_REF_TYPE or dtype == DatasetBuilder.REGION_REF_TYPE
        return False

    def __queue_ref(self, func):
        '''Set aside filling dset with references

        dest[sl] = func()

        Args:
           dset: the h5py.Dataset that the references need to be added to
           sl: the np.s_ (slice) object for indexing into dset
           func: a function to call to return the chunk of data, with
                 references filled in
        '''
        # TODO: come up with more intelligent way of
        # queueing reference resolution, based on reference
        # dependency
        self.__ref_queue.append(func)

    def __rec_get_ref(self, ref_list):
        ret = list()
        for elem in ref_list:
            if isinstance(elem, (list, tuple)):
                ret.append(self.__rec_get_ref(elem))
            elif isinstance(elem, (Builder, Container)):
                ret.append(self.__get_ref(elem))
            else:
                ret.append(elem)
        return ret

    @property
    def mode(self):
        """
        Return the HDF5 file mode. One of ("w", "r", "r+", "a", "w-", "x").
        """
        return self.__mode

    @classmethod
    @docval(*get_docval(H5DataIO.__init__))
    def set_dataio(cls, **kwargs):
        """
        Wrap the given Data object with an H5DataIO.

        This method is provided merely for convenience. It is the equivalent
        of the following:

        .. code-block:: python

            from hdmf.backends.hdf5 import H5DataIO
            data = ...
            data = H5DataIO(data)
        """
        return H5DataIO.__init__(**kwargs)



import logging
from collections import OrderedDict, deque
from copy import copy

from .builders import DatasetBuilder, GroupBuilder, LinkBuilder, Builder, BaseBuilder
from .classgenerator import ClassGenerator, CustomClassGenerator, MCIClassGenerator
from ..container import AbstractContainer, Container, Data
from ..spec import DatasetSpec, GroupSpec, NamespaceCatalog
from ..spec.spec import BaseStorageSpec
from ..utils import docval, getargs, ExtenderMeta, get_docval


class Proxy:
    """
    A temporary object to represent a Container. This gets used when resolving the true location of a
    Container's parent.
    Proxy objects allow simple bookkeeping of all potential parents a Container may have.
    This object is used by providing all the necessary information for describing the object. This object
    gets passed around and candidates are accumulated. Upon calling resolve, all saved candidates are matched
    against the information (provided to the constructor). The candidate that has an exact match is returned.
    """

    def __init__(self, manager, source, location, namespace, data_type):
        self.__source = source
        self.__location = location
        self.__namespace = namespace
        self.__data_type = data_type
        self.__manager = manager
        self.__candidates = list()

    @property
    def source(self):
        """The source of the object e.g. file source"""
        return self.__source

    @property
    def location(self):
        """The location of the object. This can be thought of as a unique path"""
        return self.__location

    @property
    def namespace(self):
        """The namespace from which the data_type of this Proxy came from"""
        return self.__namespace

    @property
    def data_type(self):
        """The data_type of Container that should match this Proxy"""
        return self.__data_type

    @docval({"name": "object", "type": (BaseBuilder, Container), "doc": "the container or builder to get a proxy for"})
    def matches(self, **kwargs):
        obj = getargs('object', kwargs)
        if not isinstance(obj, Proxy):
            obj = self.__manager.get_proxy(obj)
        return self == obj

    @docval({"name": "container", "type": Container, "doc": "the Container to add as a candidate match"})
    def add_candidate(self, **kwargs):
        container = getargs('container', kwargs)
        self.__candidates.append(container)

    def resolve(self):
        for candidate in self.__candidates:
            if self.matches(candidate):
                return candidate
        raise ValueError("No matching candidate Container found for " + self)

    def __eq__(self, other):
        return self.data_type == other.data_type and \
               self.location == other.location and \
               self.namespace == other.namespace and \
               self.source == other.source

    def __repr__(self):
        ret = dict()
        for key in ('source', 'location', 'namespace', 'data_type'):
            ret[key] = getattr(self, key, None)
        return str(ret)


class BuildManager:
    """
    A class for managing builds of AbstractContainers
    """

    def __init__(self, type_map):
        self.logger = logging.getLogger('%s.%s' % (self.__class__.__module__, self.__class__.__qualname__))
        self.__builders = dict()
        self.__containers = dict()
        self.__active_builders = set()
        self.__type_map = type_map
        self.__ref_queue = deque()  # a queue of the ReferenceBuilders that need to be added

    @property
    def namespace_catalog(self):
        return self.__type_map.namespace_catalog

    @property
    def type_map(self):
        return self.__type_map

    @docval({"name": "object", "type": (BaseBuilder, AbstractContainer),
             "doc": "the container or builder to get a proxy for"},
            {"name": "source", "type": str,
             "doc": "the source of container being built i.e. file path", 'default': None})
    def get_proxy(self, **kwargs):
        obj = getargs('object', kwargs)
        if isinstance(obj, BaseBuilder):
            return self._get_proxy_builder(obj)
        elif isinstance(obj, AbstractContainer):
            return self._get_proxy_container(obj)

    def _get_proxy_builder(self, builder):
        dt = self.__type_map.get_builder_dt(builder)
        ns = self.__type_map.get_builder_ns(builder)
        stack = list()
        tmp = builder
        while tmp is not None:
            stack.append(tmp.name)
            tmp = self.__get_parent_dt_builder(tmp)
        loc = "/".join(reversed(stack))
        return Proxy(self, builder.source, loc, ns, dt)

    def _get_proxy_container(self, container):
        ns, dt = self.__type_map.get_container_ns_dt(container)
        stack = list()
        tmp = container
        while tmp is not None:
            if isinstance(tmp, Proxy):
                stack.append(tmp.location)
                break
            else:
                stack.append(tmp.name)
                tmp = tmp.parent
        loc = "/".join(reversed(stack))
        return Proxy(self, container.container_source, loc, ns, dt)

    @docval({"name": "container", "type": AbstractContainer, "doc": "the container to convert to a Builder"},
            {"name": "source", "type": str,
             "doc": "the source of container being built i.e. file path", 'default': None},
            {"name": "spec_ext", "type": BaseStorageSpec, "doc": "a spec that further refines the base specification",
             'default': None},
            {"name": "export", "type": bool, "doc": "whether this build is for exporting",
             'default': False},
            {"name": "root", "type": bool, "doc": "whether the container is the root of the build process",
             'default': False})
    def build(self, **kwargs):
        """ Build the GroupBuilder/DatasetBuilder for the given AbstractContainer"""
        container, export = getargs('container', 'export', kwargs)
        source, spec_ext, root = getargs('source', 'spec_ext', 'root', kwargs)
        result = self.get_builder(container)
        if root:
            self.__active_builders.clear()  # reset active builders at start of build process
        if result is None:
            self.logger.debug("Building new %s '%s' (container_source: %s, source: %s, extended spec: %s, export: %s)"
                              % (container.__class__.__name__, container.name, repr(container.container_source),
                                 repr(source), spec_ext is not None, export))
            # the container_source is not set or checked when exporting
            if not export:
                if container.container_source is None:
                    container.container_source = source
                elif source is None:
                    source = container.container_source
                else:
                    if container.container_source != source:
                        raise ValueError("Cannot change container_source once set: '%s' %s.%s"
                                         % (container.name, container.__class__.__module__,
                                            container.__class__.__name__))
            # NOTE: if exporting, then existing cached builder will be ignored and overridden with new build result
            result = self.__type_map.build(container, self, source=source, spec_ext=spec_ext, export=export)
            self.prebuilt(container, result)
            self.__active_prebuilt(result)
            self.logger.debug("Done building %s '%s'" % (container.__class__.__name__, container.name))
        elif not self.__is_active_builder(result) and container.modified:
            # if builder was built on file read and is then modified (append mode), it needs to be rebuilt
            self.logger.debug("Rebuilding modified %s '%s' (source: %s, extended spec: %s)"
                              % (container.__class__.__name__, container.name,
                                 repr(source), spec_ext is not None))
            result = self.__type_map.build(container, self, builder=result, source=source, spec_ext=spec_ext,
                                           export=export)
            self.logger.debug("Done rebuilding %s '%s'" % (container.__class__.__name__, container.name))
        else:
            self.logger.debug("Using prebuilt %s '%s' for %s '%s'"
                              % (result.__class__.__name__, result.name,
                                 container.__class__.__name__, container.name))
        if root:  # create reference builders only after building all other builders
            self.__add_refs()
            self.__active_builders.clear()  # reset active builders now that build process has completed
        return result

    @docval({"name": "container", "type": AbstractContainer, "doc": "the AbstractContainer to save as prebuilt"},
            {'name': 'builder', 'type': (DatasetBuilder, GroupBuilder),
             'doc': 'the Builder representation of the given container'})
    def prebuilt(self, **kwargs):
        ''' Save the Builder for a given AbstractContainer for future use '''
        container, builder = getargs('container', 'builder', kwargs)
        container_id = self.__conthash__(container)
        self.__builders[container_id] = builder
        builder_id = self.__bldrhash__(builder)
        self.__containers[builder_id] = container

    def __active_prebuilt(self, builder):
        """Save the Builder for future use during the active/current build process."""
        builder_id = self.__bldrhash__(builder)
        self.__active_builders.add(builder_id)

    def __is_active_builder(self, builder):
        """Return True if the Builder was created during the active/current build process."""
        builder_id = self.__bldrhash__(builder)
        return builder_id in self.__active_builders

    def __conthash__(self, obj):
        return id(obj)

    def __bldrhash__(self, obj):
        return id(obj)

    def __add_refs(self):
        '''
        Add ReferenceBuilders.

        References get queued to be added after all other objects are built. This is because
        the current traversal algorithm (i.e. iterating over specs)
        does not happen in a guaranteed order. We need to build the targets
        of the reference builders so that the targets have the proper parent,
        and then write the reference builders after we write everything else.
        '''
        while len(self.__ref_queue) > 0:
            call = self.__ref_queue.popleft()
            self.logger.debug("Adding ReferenceBuilder with call id %d from queue (length %d)"
                              % (id(call), len(self.__ref_queue)))
            call()

    def queue_ref(self, func):
        '''Set aside creating ReferenceBuilders'''
        # TODO: come up with more intelligent way of
        # queueing reference resolution, based on reference
        # dependency
        self.__ref_queue.append(func)

    def purge_outdated(self):
        containers_copy = self.__containers.copy()
        for container in containers_copy.values():
            if container.modified:
                container_id = self.__conthash__(container)
                builder = self.__builders.get(container_id)
                builder_id = self.__bldrhash__(builder)
                self.logger.debug("Purging %s '%s' for %s '%s' from prebuilt cache"
                                  % (builder.__class__.__name__, builder.name,
                                     container.__class__.__name__, container.name))
                self.__builders.pop(container_id)
                self.__containers.pop(builder_id)

    @docval({"name": "container", "type": AbstractContainer, "doc": "the container to get the builder for"})
    def get_builder(self, **kwargs):
        """Return the prebuilt builder for the given container or None if it does not exist."""
        container = getargs('container', kwargs)
        container_id = self.__conthash__(container)
        result = self.__builders.get(container_id)
        return result

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder),
             'doc': 'the builder to construct the AbstractContainer from'})
    def construct(self, **kwargs):
        """ Construct the AbstractContainer represented by the given builder """
        builder = getargs('builder', kwargs)
        if isinstance(builder, LinkBuilder):
            builder = builder.target
        builder_id = self.__bldrhash__(builder)
        result = self.__containers.get(builder_id)
        if result is None:
            parent_builder = self.__get_parent_dt_builder(builder)
            if parent_builder is not None:
                parent = self._get_proxy_builder(parent_builder)
                result = self.__type_map.construct(builder, self, parent)
            else:
                # we are at the top of the hierarchy,
                # so it must be time to resolve parents
                result = self.__type_map.construct(builder, self, None)
                self.__resolve_parents(result)
            self.prebuilt(result, builder)
        result.set_modified(False)
        return result

    def __resolve_parents(self, container):
        stack = [container]
        while len(stack) > 0:
            tmp = stack.pop()
            if isinstance(tmp.parent, Proxy):
                tmp.parent = tmp.parent.resolve()
            for child in tmp.children:
                stack.append(child)

    def __get_parent_dt_builder(self, builder):
        '''
        Get the next builder above the given builder
        that has a data_type
        '''
        tmp = builder.parent
        ret = None
        while tmp is not None:
            ret = tmp
            dt = self.__type_map.get_builder_dt(tmp)
            if dt is not None:
                break
            tmp = tmp.parent
        return ret

    # *** The following methods just delegate calls to self.__type_map ***

    @docval({'name': 'builder', 'type': Builder, 'doc': 'the Builder to get the class object for'})
    def get_cls(self, **kwargs):
        ''' Get the class object for the given Builder '''
        builder = getargs('builder', kwargs)
        return self.__type_map.get_cls(builder)

    @docval({"name": "container", "type": AbstractContainer, "doc": "the container to convert to a Builder"},
            returns='The name a Builder should be given when building this container', rtype=str)
    def get_builder_name(self, **kwargs):
        ''' Get the name a Builder should be given '''
        container = getargs('container', kwargs)
        return self.__type_map.get_builder_name(container)

    @docval({'name': 'spec', 'type': (DatasetSpec, GroupSpec), 'doc': 'the parent spec to search'},
            {'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the sub-specification for'})
    def get_subspec(self, **kwargs):
        ''' Get the specification from this spec that corresponds to the given builder '''
        spec, builder = getargs('spec', 'builder', kwargs)
        return self.__type_map.get_subspec(spec, builder)

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the sub-specification for'})
    def get_builder_ns(self, **kwargs):
        ''' Get the namespace of a builder '''
        builder = getargs('builder', kwargs)
        return self.__type_map.get_builder_ns(builder)

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the data_type for'})
    def get_builder_dt(self, **kwargs):
        '''
        Get the data_type of a builder
        '''
        builder = getargs('builder', kwargs)
        return self.__type_map.get_builder_dt(builder)

    @docval({'name': 'builder', 'type': (GroupBuilder, DatasetBuilder, AbstractContainer),
             'doc': 'the builder or container to check'},
            {'name': 'parent_data_type', 'type': str,
             'doc': 'the potential parent data_type that refers to a data_type'},
            returns="True if data_type of *builder* is a sub-data_type of *parent_data_type*, False otherwise",
            rtype=bool)
    def is_sub_data_type(self, **kwargs):
        '''
        Return whether or not data_type of *builder* is a sub-data_type of *parent_data_type*
        '''
        builder, parent_dt = getargs('builder', 'parent_data_type', kwargs)
        if isinstance(builder, (GroupBuilder, DatasetBuilder)):
            ns = self.get_builder_ns(builder)
            dt = self.get_builder_dt(builder)
        else:  # builder is an AbstractContainer
            ns, dt = self.type_map.get_container_ns_dt(builder)
        return self.namespace_catalog.is_sub_data_type(ns, dt, parent_dt)


class TypeSource:
    '''A class to indicate the source of a data_type in a namespace.
    This class should only be used by TypeMap
    '''

    @docval({"name": "namespace", "type": str, "doc": "the namespace the from, which the data_type originated"},
            {"name": "data_type", "type": str, "doc": "the name of the type"})
    def __init__(self, **kwargs):
        namespace, data_type = getargs('namespace', 'data_type', kwargs)
        self.__namespace = namespace
        self.__data_type = data_type

    @property
    def namespace(self):
        return self.__namespace

    @property
    def data_type(self):
        return self.__data_type


class TypeMap:
    ''' A class to maintain the map between ObjectMappers and AbstractContainer classes
    '''

    @docval({'name': 'namespaces', 'type': NamespaceCatalog, 'doc': 'the NamespaceCatalog to use', 'default': None},
            {'name': 'mapper_cls', 'type': type, 'doc': 'the ObjectMapper class to use', 'default': None})
    def __init__(self, **kwargs):
        namespaces, mapper_cls = getargs('namespaces', 'mapper_cls', kwargs)
        if namespaces is None:
            namespaces = NamespaceCatalog()
        if mapper_cls is None:
            from .objectmapper import ObjectMapper  # avoid circular import
            mapper_cls = ObjectMapper
        self.__ns_catalog = namespaces
        self.__mappers = dict()  # already constructed ObjectMapper classes
        self.__mapper_cls = dict()  # the ObjectMapper class to use for each container type
        self.__container_types = OrderedDict()
        self.__data_types = dict()
        self.__default_mapper_cls = mapper_cls
        self.__class_generator = ClassGenerator()
        self.register_generator(CustomClassGenerator)
        self.register_generator(MCIClassGenerator)

    @property
    def namespace_catalog(self):
        return self.__ns_catalog

    @property
    def container_types(self):
        return self.__container_types

    def __copy__(self):
        ret = TypeMap(copy(self.__ns_catalog), self.__default_mapper_cls)
        ret.merge(self)
        return ret

    def __deepcopy__(self, memo):
        # XXX: From @nicain: All of a sudden legacy tests started
        #      needing this argument in deepcopy. Doesn't hurt anything, though.
        return self.__copy__()

    def copy_mappers(self, type_map):
        for namespace in self.__ns_catalog.namespaces:
            if namespace not in type_map.__container_types:
                continue
            for data_type in self.__ns_catalog.get_namespace(namespace).get_registered_types():
                container_cls = type_map.__container_types[namespace].get(data_type)
                if container_cls is None:
                    continue
                self.register_container_type(namespace, data_type, container_cls)
                if container_cls in type_map.__mapper_cls:
                    self.register_map(container_cls, type_map.__mapper_cls[container_cls])

    def merge(self, type_map, ns_catalog=False):
        if ns_catalog:
            self.namespace_catalog.merge(type_map.namespace_catalog)
        for namespace in type_map.__container_types:
            for data_type in type_map.__container_types[namespace]:
                container_cls = type_map.__container_types[namespace][data_type]
                self.register_container_type(namespace, data_type, container_cls)
        for container_cls in type_map.__mapper_cls:
            self.register_map(container_cls, type_map.__mapper_cls[container_cls])
        for custom_generators in reversed(type_map.__class_generator.custom_generators):
            # iterate in reverse order because generators are stored internally as a stack
            self.register_generator(custom_generators)

    @docval({"name": "generator", "type": type, "doc": "the CustomClassGenerator class to register"})
    def register_generator(self, **kwargs):
        """Add a custom class generator."""
        generator = getargs('generator', kwargs)
        self.__class_generator.register_generator(generator)

    @docval(*get_docval(NamespaceCatalog.load_namespaces),
            returns="the namespaces loaded from the given file", rtype=dict)
    def load_namespaces(self, **kwargs):
        '''Load namespaces from a namespace file.
        This method will call load_namespaces on the NamespaceCatalog used to construct this TypeMap. Additionally,
        it will process the return value to keep track of what types were included in the loaded namespaces. Calling
        load_namespaces here has the advantage of being able to keep track of type dependencies across namespaces.
        '''
        deps = self.__ns_catalog.load_namespaces(**kwargs)
        for new_ns, ns_deps in deps.items():
            for src_ns, types in ns_deps.items():
                for dt in types:
                    container_cls = self.get_dt_container_cls(dt, src_ns, autogen=False)
                    if container_cls is None:
                        container_cls = TypeSource(src_ns, dt)
                    self.register_container_type(new_ns, dt, container_cls)
        return deps

    @docval({"name": "namespace", "type": str, "doc": "the namespace containing the data_type"},
            {"name": "data_type", "type": str, "doc": "the data type to create a AbstractContainer class for"},
            {"name": "autogen", "type": bool, "doc": "autogenerate class if one does not exist", "default": True},
            returns='the class for the given namespace and data_type', rtype=type)
    def get_container_cls(self, **kwargs):
        """Get the container class from data type specification.
        If no class has been associated with the ``data_type`` from ``namespace``, a class will be dynamically
        created and returned.
        """
        # NOTE: this internally used function get_container_cls will be removed in favor of get_dt_container_cls
        namespace, data_type, autogen = getargs('namespace', 'data_type', 'autogen', kwargs)
        return self.get_dt_container_cls(data_type, namespace, autogen)

    @docval({"name": "data_type", "type": str, "doc": "the data type to create a AbstractContainer class for"},
            {"name": "namespace", "type": str, "doc": "the namespace containing the data_type", "default": None},
            {"name": "autogen", "type": bool, "doc": "autogenerate class if one does not exist", "default": True},
            returns='the class for the given namespace and data_type', rtype=type)
    def get_dt_container_cls(self, **kwargs):
        """Get the container class from data type specification.
        If no class has been associated with the ``data_type`` from ``namespace``, a class will be dynamically
        created and returned.

        Replaces get_container_cls but namespace is optional. If namespace is unknown, it will be looked up from
        all namespaces.
        """
        namespace, data_type, autogen = getargs('namespace', 'data_type', 'autogen', kwargs)

        # namespace is unknown, so look it up
        if namespace is None:
            for ns_key, ns_data_types in self.__container_types.items():
                # NOTE that the type_name may appear in multiple namespaces based on how they were resolved
                # but the same type_name should point to the same class
                if data_type in ns_data_types:
                    namespace = ns_key
                    break

        cls = self.__get_container_cls(namespace, data_type)
        if cls is None and autogen:  # dynamically generate a class
            spec = self.__ns_catalog.get_spec(namespace, data_type)
            self.__check_dependent_types(spec, namespace)
            parent_cls = self.__get_parent_cls(namespace, data_type, spec)
            attr_names = self.__default_mapper_cls.get_attr_names(spec)
            cls = self.__class_generator.generate_class(data_type, spec, parent_cls, attr_names, self)
            self.register_container_type(namespace, data_type, cls)
        return cls

    def __check_dependent_types(self, spec, namespace):
        """Ensure that classes for all types used by this type exist in this namespace and generate them if not.
        """
        def __check_dependent_types_helper(spec, namespace):
            if isinstance(spec, (GroupSpec, DatasetSpec)):
                if spec.data_type_inc is not None:
                    self.get_dt_container_cls(spec.data_type_inc, namespace)  # TODO handle recursive definitions
                if spec.data_type_def is not None:  # nested type definition
                    self.get_dt_container_cls(spec.data_type_def, namespace)
            else:  # spec is a LinkSpec
                self.get_dt_container_cls(spec.target_type, namespace)
            if isinstance(spec, GroupSpec):
                for child_spec in (spec.groups + spec.datasets + spec.links):
                    __check_dependent_types_helper(child_spec, namespace)

        if spec.data_type_inc is not None:
            self.get_dt_container_cls(spec.data_type_inc, namespace)
        if isinstance(spec, GroupSpec):
            for child_spec in (spec.groups + spec.datasets + spec.links):
                __check_dependent_types_helper(child_spec, namespace)

    def __get_parent_cls(self, namespace, data_type, spec):
        dt_hier = self.__ns_catalog.get_hierarchy(namespace, data_type)
        dt_hier = dt_hier[1:]  # remove the current data_type
        parent_cls = None
        for t in dt_hier:
            parent_cls = self.__get_container_cls(namespace, t)
            if parent_cls is not None:
                break
        if parent_cls is None:
            if isinstance(spec, GroupSpec):
                parent_cls = Container
            elif isinstance(spec, DatasetSpec):
                parent_cls = Data
            else:
                raise ValueError("Cannot generate class from %s" % type(spec))
        if type(parent_cls) is not ExtenderMeta:
            raise ValueError("parent class %s is not of type ExtenderMeta - %s" % (parent_cls, type(parent_cls)))
        return parent_cls

    def __get_container_cls(self, namespace, data_type):
        """Get the container class for the namespace, data_type. If the class doesn't exist yet, generate it."""
        if namespace not in self.__container_types:
            return None
        if data_type not in self.__container_types[namespace]:
            return None
        ret = self.__container_types[namespace][data_type]
        if isinstance(ret, TypeSource):  # data_type is a dependency from ret.namespace
            cls = self.get_dt_container_cls(ret.data_type, ret.namespace)  # get class / generate class
            # register the same class into this namespace (replaces TypeSource)
            self.register_container_type(namespace, data_type, cls)
            ret = cls
        return ret

    @docval({'name': 'obj', 'type': (GroupBuilder, DatasetBuilder, LinkBuilder, GroupSpec, DatasetSpec),
             'doc': 'the object to get the type key for'})
    def __type_key(self, obj):
        """
        A wrapper function to simplify the process of getting a type_key for an object.
        The type_key is used to get the data_type from a Builder's attributes.
        """
        if isinstance(obj, LinkBuilder):
            obj = obj.builder
        if isinstance(obj, (GroupBuilder, GroupSpec)):
            return self.__ns_catalog.group_spec_cls.type_key()
        else:
            return self.__ns_catalog.dataset_spec_cls.type_key()

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the data_type for'})
    def get_builder_dt(self, **kwargs):
        '''
        Get the data_type of a builder
        '''
        builder = getargs('builder', kwargs)
        ret = None
        if isinstance(builder, LinkBuilder):
            builder = builder.builder
        if isinstance(builder, GroupBuilder):
            ret = builder.attributes.get(self.__ns_catalog.group_spec_cls.type_key())
        else:
            ret = builder.attributes.get(self.__ns_catalog.dataset_spec_cls.type_key())
        if isinstance(ret, bytes):
            ret = ret.decode('UTF-8')
        return ret

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the sub-specification for'})
    def get_builder_ns(self, **kwargs):
        ''' Get the namespace of a builder '''
        builder = getargs('builder', kwargs)
        if isinstance(builder, LinkBuilder):
            builder = builder.builder
        ret = builder.attributes.get('namespace')
        return ret

    @docval({'name': 'builder', 'type': Builder,
             'doc': 'the Builder object to get the corresponding AbstractContainer class for'})
    def get_cls(self, **kwargs):
        ''' Get the class object for the given Builder '''
        builder = getargs('builder', kwargs)
        data_type = self.get_builder_dt(builder)
        if data_type is None:
            raise ValueError("No data_type found for builder %s" % builder.path)
        namespace = self.get_builder_ns(builder)
        if namespace is None:
            raise ValueError("No namespace found for builder %s" % builder.path)
        return self.get_dt_container_cls(data_type, namespace)

    @docval({'name': 'spec', 'type': (DatasetSpec, GroupSpec), 'doc': 'the parent spec to search'},
            {'name': 'builder', 'type': (DatasetBuilder, GroupBuilder, LinkBuilder),
             'doc': 'the builder to get the sub-specification for'})
    def get_subspec(self, **kwargs):
        ''' Get the specification from this spec that corresponds to the given builder '''
        spec, builder = getargs('spec', 'builder', kwargs)
        if isinstance(builder, LinkBuilder):
            builder_type = type(builder.builder)
            # TODO consider checking against spec.get_link
        else:
            builder_type = type(builder)
        if issubclass(builder_type, DatasetBuilder):
            subspec = spec.get_dataset(builder.name)
        else:
            subspec = spec.get_group(builder.name)
        if subspec is None:
            # builder was generated from something with a data_type and a wildcard name
            if isinstance(builder, LinkBuilder):
                dt = self.get_builder_dt(builder.builder)
            else:
                dt = self.get_builder_dt(builder)
            if dt is not None:
                ns = self.get_builder_ns(builder)
                hierarchy = self.__ns_catalog.get_hierarchy(ns, dt)
                for t in hierarchy:
                    subspec = spec.get_data_type(t)
                    if subspec is not None:
                        break
                    subspec = spec.get_target_type(t)
                    if subspec is not None:
                        break
        return subspec

    def get_container_ns_dt(self, obj):
        container_cls = obj.__class__
        namespace, data_type = self.get_container_cls_dt(container_cls)
        return namespace, data_type

    def get_container_cls_dt(self, cls):
        def_ret = (None, None)
        for _cls in cls.__mro__:  # pragma: no branch
            ret = self.__data_types.get(_cls, def_ret)
            if ret is not def_ret:
                return ret
        return ret

    @docval({'name': 'namespace', 'type': str,
             'doc': 'the namespace to get the container classes for', 'default': None})
    def get_container_classes(self, **kwargs):
        namespace = getargs('namespace', kwargs)
        ret = self.__data_types.keys()
        if namespace is not None:
            ret = filter(lambda x: self.__data_types[x][0] == namespace, ret)
        return list(ret)

    @docval({'name': 'obj', 'type': (AbstractContainer, Builder), 'doc': 'the object to get the ObjectMapper for'},
            returns='the ObjectMapper to use for mapping the given object', rtype='ObjectMapper')
    def get_map(self, **kwargs):
        """ Return the ObjectMapper object that should be used for the given container """
        obj = getargs('obj', kwargs)
        # get the container class, and namespace/data_type
        if isinstance(obj, AbstractContainer):
            container_cls = obj.__class__
            namespace, data_type = self.get_container_cls_dt(container_cls)
            if namespace is None:
                raise ValueError("class %s is not mapped to a data_type" % container_cls)
        else:
            data_type = self.get_builder_dt(obj)
            namespace = self.get_builder_ns(obj)
            container_cls = self.get_cls(obj)
        # now build the ObjectMapper class
        mapper = self.__mappers.get(container_cls)
        if mapper is None:
            mapper_cls = self.__default_mapper_cls
            for cls in container_cls.__mro__:
                tmp_mapper_cls = self.__mapper_cls.get(cls)
                if tmp_mapper_cls is not None:
                    mapper_cls = tmp_mapper_cls
                    break
            spec = self.__ns_catalog.get_spec(namespace, data_type)
            mapper = mapper_cls(spec)
            self.__mappers[container_cls] = mapper
        return mapper

    @docval({"name": "namespace", "type": str, "doc": "the namespace containing the data_type to map the class to"},
            {"name": "data_type", "type": str, "doc": "the data_type to map the class to"},
            {"name": "container_cls", "type": (TypeSource, type), "doc": "the class to map to the specified data_type"})
    def register_container_type(self, **kwargs):
        ''' Map a container class to a data_type '''
        namespace, data_type, container_cls = getargs('namespace', 'data_type', 'container_cls', kwargs)
        spec = self.__ns_catalog.get_spec(namespace, data_type)  # make sure the spec exists
        self.__container_types.setdefault(namespace, dict())
        self.__container_types[namespace][data_type] = container_cls
        self.__data_types.setdefault(container_cls, (namespace, data_type))
        if not isinstance(container_cls, TypeSource):
            setattr(container_cls, spec.type_key(), data_type)
            setattr(container_cls, 'namespace', namespace)

    @docval({"name": "container_cls", "type": type,
             "doc": "the AbstractContainer class for which the given ObjectMapper class gets used for"},
            {"name": "mapper_cls", "type": type, "doc": "the ObjectMapper class to use to map"})
    def register_map(self, **kwargs):
        ''' Map a container class to an ObjectMapper class '''
        container_cls, mapper_cls = getargs('container_cls', 'mapper_cls', kwargs)
        if self.get_container_cls_dt(container_cls) == (None, None):
            raise ValueError('cannot register map for type %s - no data_type found' % container_cls)
        self.__mapper_cls[container_cls] = mapper_cls

    @docval({"name": "container", "type": AbstractContainer, "doc": "the container to convert to a Builder"},
            {"name": "manager", "type": BuildManager,
             "doc": "the BuildManager to use for managing this build", 'default': None},
            {"name": "source", "type": str,
             "doc": "the source of container being built i.e. file path", 'default': None},
            {"name": "builder", "type": BaseBuilder, "doc": "the Builder to build on", 'default': None},
            {"name": "spec_ext", "type": BaseStorageSpec, "doc": "a spec extension", 'default': None},
            {"name": "export", "type": bool, "doc": "whether this build is for exporting",
             'default': False})
    def build(self, **kwargs):
        """Build the GroupBuilder/DatasetBuilder for the given AbstractContainer"""
        container, manager, builder = getargs('container', 'manager', 'builder', kwargs)
        source, spec_ext, export = getargs('source', 'spec_ext', 'export', kwargs)

        # get the ObjectMapper to map between Spec objects and AbstractContainer attributes
        obj_mapper = self.get_map(container)
        if obj_mapper is None:
            raise ValueError('No ObjectMapper found for container of type %s' % str(container.__class__.__name__))

        # convert the container to a builder using the ObjectMapper
        if manager is None:
            manager = BuildManager(self)
        builder = obj_mapper.build(container, manager, builder=builder, source=source, spec_ext=spec_ext, export=export)

        # add additional attributes (namespace, data_type, object_id) to builder
        namespace, data_type = self.get_container_ns_dt(container)
        builder.set_attribute('namespace', namespace)
        builder.set_attribute(self.__type_key(obj_mapper.spec), data_type)
        builder.set_attribute(obj_mapper.spec.id_key(), container.object_id)
        return builder

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder),
             'doc': 'the builder to construct the AbstractContainer from'},
            {'name': 'build_manager', 'type': BuildManager,
             'doc': 'the BuildManager for constructing', 'default': None},
            {'name': 'parent', 'type': (Proxy, Container),
             'doc': 'the parent Container/Proxy for the Container being built', 'default': None})
    def construct(self, **kwargs):
        """ Construct the AbstractContainer represented by the given builder """
        builder, build_manager, parent = getargs('builder', 'build_manager', 'parent', kwargs)
        if build_manager is None:
            build_manager = BuildManager(self)
        obj_mapper = self.get_map(builder)
        if obj_mapper is None:
            dt = builder.attributes[self.namespace_catalog.group_spec_cls.type_key()]
            raise ValueError('No ObjectMapper found for builder of type %s' % dt)
        else:
            return obj_mapper.construct(builder, build_manager, parent)

    @docval({"name": "container", "type": AbstractContainer, "doc": "the container to convert to a Builder"},
            returns='The name a Builder should be given when building this container', rtype=str)
    def get_builder_name(self, **kwargs):
        ''' Get the name a Builder should be given '''
        container = getargs('container', kwargs)
        obj_mapper = self.get_map(container)
        if obj_mapper is None:
            raise ValueError('No ObjectMapper found for container of type %s' % str(container.__class__.__name__))
        else:
            return obj_mapper.get_builder_name(container)



import unittest
import numpy as np
from pathlib import Path
from tempfile import mkdtemp
from shutil import rmtree
from typing import Tuple, Iterable

import h5py

from hdmf.data_utils import GenericDataChunkIterator
from hdmf.testing import TestCase

try:
    import tqdm  # noqa: F401
    TQDM_INSTALLED = True
except ImportError:
    TQDM_INSTALLED = False


class GenericDataChunkIteratorTests(TestCase):
    class TestNumpyArrayDataChunkIterator(GenericDataChunkIterator):
        def __init__(self, array: np.ndarray, **kwargs):
            self.array = array
            super().__init__(**kwargs)

        def _get_data(self, selection) -> np.ndarray:
            return self.array[selection]

        def _get_maxshape(self) -> Tuple[int, ...]:
            return self.array.shape

        def _get_dtype(self) -> np.dtype:
            return self.array.dtype

    class TestNumpyArrayDataChunkIteratorWithNumpyDtypeShape(GenericDataChunkIterator):
        def __init__(self, array: np.ndarray, **kwargs):
            self.array = array
            super().__init__(**kwargs)

        def _get_data(self, selection) -> np.ndarray:
            return self.array[selection]

        def _get_maxshape(self) -> Tuple[np.uint64, ...]:  # Undesirable return type, but can be handled
            return tuple(np.uint64(x) for x in self.array.shape)

        def _get_dtype(self) -> np.dtype:
            return self.array.dtype

    @classmethod
    def setUpClass(cls):
        cls.test_dir = Path(mkdtemp())
        cls.test_array = np.empty(shape=(2000, 384), dtype="int16")

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.test_dir)

    def check_first_data_chunk_call(self, expected_selection, iterator_options):
        test = self.TestNumpyArrayDataChunkIterator(array=self.test_array, **iterator_options)
        first_data_chunk = next(test)
        self.assertEqual(first_data_chunk.selection, expected_selection)
        np.testing.assert_array_equal(first_data_chunk, self.test_array[expected_selection])

    def check_direct_hdf5_write(self, iterator_options):
        iterator = self.TestNumpyArrayDataChunkIterator(
            array=self.test_array, **iterator_options
        )
        with h5py.File(name=self.test_dir / "test_generic_iterator_array.hdf5", mode="w") as f:
            dset = f.create_dataset(
                name="test", shape=self.test_array.shape, dtype="int16", chunks=iterator.chunk_shape
            )
            for chunk in iterator:
                dset[chunk.selection] = chunk.data
            np.testing.assert_array_equal(np.array(dset), self.test_array)
            self.assertEqual(dset.chunks, iterator.chunk_shape)

    def check_all_of_iterable_is_python_int(self, iterable: Iterable):
        assert all(
            tuple(  # Easier to visualize failures in pytest with tuple vs. generator
                isinstance(x, int) for x in iterable
            )
        )

    def test_abstract_assertions(self):
        class TestGenericDataChunkIterator(GenericDataChunkIterator):
            pass

        with self.assertRaisesWith(
            exc_type=TypeError,
            exc_msg=(
                "Can't instantiate abstract class TestGenericDataChunkIterator with abstract methods "
                "_get_data, _get_dtype, _get_maxshape"
            ),
        ):
            TestGenericDataChunkIterator()

    def test_joint_option_assertions(self):
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg="Only one of 'buffer_gb' or 'buffer_shape' can be specified!",
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, buffer_shape=(2000, 384), buffer_gb=1)

        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg="Only one of 'chunk_mb' or 'chunk_shape' can be specified!",
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_shape=(1580, 316), chunk_mb=1)

        chunk_shape = (2001, 384)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=(
                f"Some dimensions of chunk_shape ({chunk_shape}) exceed the "
                f"data dimensions ((2000, 384))!"
            ),
        ):
            self.TestNumpyArrayDataChunkIterator(
                array=self.test_array, chunk_shape=chunk_shape
            )

        buffer_shape = (1000, 192)
        chunk_shape = (100, 384)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=(
                f"Some dimensions of chunk_shape ({chunk_shape}) exceed the "
                f"buffer shape ({buffer_shape})!"
            ),
        ):
            self.TestNumpyArrayDataChunkIterator(
                array=self.test_array, buffer_shape=buffer_shape, chunk_shape=chunk_shape
            )

        buffer_shape = (1000, 192)
        chunk_shape = (1000, 5)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=(
                f"Some dimensions of chunk_shape ({chunk_shape}) do not evenly divide the "
                f"buffer shape ({buffer_shape})!"
            ),
        ):
            self.TestNumpyArrayDataChunkIterator(
                array=self.test_array, buffer_shape=buffer_shape, chunk_shape=chunk_shape
            )

    def test_buffer_option_assertion_negative_buffer_gb(self):
        buffer_gb = -1
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"buffer_gb ({buffer_gb}) must be greater than zero!"
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, buffer_gb=buffer_gb)

    def test_buffer_option_assertion_exceed_maxshape(self):
        buffer_shape = (2001, 384)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=(
                f"Some dimensions of buffer_shape ({buffer_shape}) exceed the data "
                f"dimensions ({self.test_array.shape})!"
            )
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, buffer_shape=buffer_shape)

    def test_buffer_option_assertion_negative_shape(self):
        buffer_shape = (-1, 384)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"Some dimensions of buffer_shape ({buffer_shape}) are less than zero!"
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, buffer_shape=buffer_shape)

    def test_chunk_option_assertion_negative_chunk_mb(self):
        chunk_mb = -1
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"chunk_mb ({chunk_mb}) must be greater than zero!"
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_mb=chunk_mb)

    def test_chunk_option_assertion_negative_shape(self):
        chunk_shape = (-1, 384)
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"Some dimensions of chunk_shape ({chunk_shape}) are less than zero!"
        ):
            self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_shape=chunk_shape)

    @unittest.skipIf(not TQDM_INSTALLED, "optional tqdm module is not installed")
    def test_progress_bar_assertion(self):
        with self.assertWarnsWith(
            warn_type=UserWarning,
            exc_msg="Option 'total' in 'progress_bar_options' is not allowed to be over-written! Ignoring."
        ):
            _ = self.TestNumpyArrayDataChunkIterator(
                array=self.test_array,
                display_progress=True,
                progress_bar_options=dict(total=5),
            )

    def test_maxshape_attribute_contains_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIterator(array=self.test_array).maxshape
        )

    def test_automated_buffer_shape_attribute_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIterator(array=self.test_array).buffer_shape
        )

    def test_automated_chunk_shape_attribute_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIterator(array=self.test_array).chunk_shape
        )

    def test_np_dtype_maxshape_attribute_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIteratorWithNumpyDtypeShape(array=self.test_array).maxshape
        )

    def test_manual_buffer_shape_attribute_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIterator(
                array=self.test_array,
                chunk_shape=(np.uint64(100), np.uint64(2)),
                buffer_shape=(np.uint64(200), np.uint64(4)),
            ).buffer_shape
        )

    def test_manual_chunk_shape_attribute_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        self.check_all_of_iterable_is_python_int(
            iterable=self.TestNumpyArrayDataChunkIterator(
                array=self.test_array,
                chunk_shape=(np.uint64(100), np.uint64(2))
            ).chunk_shape
        )

    def test_selection_slices_int_type(self):
        """Motivated by issues described in https://github.com/hdmf-dev/hdmf/pull/780 & 781 regarding return types."""
        iterator = self.TestNumpyArrayDataChunkIterator(array=self.test_array)
        first_chunk = next(iterator)
        stop_0 = first_chunk.selection[0].stop
        start_0 = first_chunk.selection[0].start
        stop_1 = first_chunk.selection[1].stop
        start_1 = first_chunk.selection[1].start

        self.check_all_of_iterable_is_python_int(iterable=(stop_0, start_0, stop_1, start_1))

    def test_num_buffers(self):
        buffer_shape = (950, 190)
        chunk_shape = (50, 38)
        expected_num_buffers = 9

        test = self.TestNumpyArrayDataChunkIterator(
            array=self.test_array, buffer_shape=buffer_shape, chunk_shape=chunk_shape
        )
        self.assertEqual(first=test.num_buffers, second=expected_num_buffers)

    def test_numpy_array_chunk_iterator(self):
        iterator_options = dict()
        self.check_first_data_chunk_call(
            expected_selection=(slice(0, 2000), slice(0, 384)), iterator_options=iterator_options
        )
        self.check_direct_hdf5_write(iterator_options=iterator_options)

    def test_buffer_shape_option(self):
        expected_buffer_shape = (1580, 316)
        iterator_options = dict(buffer_shape=expected_buffer_shape)
        self.check_first_data_chunk_call(
            expected_selection=tuple([slice(0, buffer_shape_axis) for buffer_shape_axis in expected_buffer_shape]),
            iterator_options=iterator_options,
        )
        self.check_direct_hdf5_write(iterator_options=iterator_options)

    def test_buffer_gb_option(self):
        # buffer is smaller than default chunk; should collapse to chunk shape
        resulting_buffer_shape = (1580, 316)
        iterator_options = dict(buffer_gb=0.0005)
        self.check_first_data_chunk_call(
            expected_selection=tuple(
                [
                    slice(0, buffer_shape_axis)
                    for buffer_shape_axis in resulting_buffer_shape
                ]
            ),
            iterator_options=iterator_options,
        )
        self.check_direct_hdf5_write(iterator_options=iterator_options)

        # buffer is larger than total data size; should collapse to maxshape
        resulting_buffer_shape = (2000, 384)
        for buffer_gb_input_dtype_pass in [2, 2.0]:
            iterator_options = dict(buffer_gb=2)
            self.check_first_data_chunk_call(
                expected_selection=tuple(
                    [
                        slice(0, buffer_shape_axis)
                        for buffer_shape_axis in resulting_buffer_shape
                    ]
                ),
                iterator_options=iterator_options,
            )
            self.check_direct_hdf5_write(iterator_options=iterator_options)

    def test_chunk_shape_option(self):
        test_chunk_shape = (1580, 316)
        iterator = self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_shape=test_chunk_shape)
        self.assertEqual(iterator.chunk_shape, test_chunk_shape)

    def test_chunk_mb_option(self):
        expected_chunk_shape = (1115, 223)
        iterator = self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_mb=0.5)
        self.assertEqual(iterator.chunk_shape, expected_chunk_shape)

    def test_chunk_mb_option_larger_than_total_size(self):
        """Chunk is larger than total data size; should collapse to maxshape."""
        expected_chunk_shape = (2000, 384)
        iterator = self.TestNumpyArrayDataChunkIterator(array=self.test_array, chunk_mb=2)
        self.assertEqual(iterator.chunk_shape, expected_chunk_shape)

    def test_chunk_mb_option_while_condition(self):
        """Test to evoke while condition of default shaping method."""
        expected_chunk_shape = (2, 79, 79)
        special_array = np.random.randint(low=-(2 ** 15), high=2 ** 15 - 1, size=(2, 2000, 2000), dtype="int16")
        iterator = self.TestNumpyArrayDataChunkIterator(array=special_array)
        self.assertEqual(iterator.chunk_shape, expected_chunk_shape)

    def test_chunk_mb_option_while_condition_unit_maxshape_axis(self):
        """Test to evoke while condition of default shaping method."""
        expected_chunk_shape = (1, 79, 79)
        special_array = np.random.randint(low=-(2 ** 15), high=2 ** 15 - 1, size=(1, 2000, 2000), dtype="int16")
        iterator = self.TestNumpyArrayDataChunkIterator(array=special_array)
        self.assertEqual(iterator.chunk_shape, expected_chunk_shape)

    @unittest.skipIf(not TQDM_INSTALLED, "optional tqdm module is not installed")
    def test_progress_bar(self):
        out_text_file = self.test_dir / "test_progress_bar.txt"
        desc = "Testing progress bar..."
        with open(file=out_text_file, mode="w") as file:
            iterator = self.TestNumpyArrayDataChunkIterator(
                array=self.test_array, display_progress=True, progress_bar_options=dict(file=file, desc=desc)
            )
            j = 0
            for buffer in iterator:
                j += 1  # dummy operation; must be silent for proper updating of bar
        with open(file=out_text_file, mode="r") as file:
            first_line = file.read()
            self.assertIn(member=desc, container=first_line)

    @unittest.skipIf(not TQDM_INSTALLED, "optional tqdm module is installed")
    def test_progress_bar_no_options(self):
        dci = self.TestNumpyArrayDataChunkIterator(array=self.test_array, display_progress=True)
        self.assertIsNotNone(dci.progress_bar)
        self.assertTrue(dci.display_progress)

    @unittest.skipIf(TQDM_INSTALLED, "optional tqdm module is not installed")
    def test_tqdm_not_installed(self):
        with self.assertWarnsWith(
            warn_type=UserWarning,
            exc_msg=("You must install tqdm to use the progress bar feature (pip install tqdm)! "
                     "Progress bar is disabled.")
        ):
            dci = self.TestNumpyArrayDataChunkIterator(
                array=self.test_array,
                display_progress=True,
            )
            self.assertFalse(dci.display_progress)



import copy as _copy
import itertools as _itertools
import posixpath as _posixpath
from abc import ABCMeta
from collections.abc import Iterable
from datetime import datetime

import numpy as np
from h5py import RegionReference

from ..utils import docval, getargs, get_docval


class Builder(dict, metaclass=ABCMeta):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the group'},
            {'name': 'parent', 'type': 'Builder', 'doc': 'the parent builder of this Builder', 'default': None},
            {'name': 'source', 'type': str,
             'doc': 'the source of the data in this builder e.g. file name', 'default': None})
    def __init__(self, **kwargs):
        name, parent, source = getargs('name', 'parent', 'source', kwargs)
        super().__init__()
        self.__name = name
        self.__parent = parent
        if source is not None:
            self.__source = source
        elif parent is not None:
            self.__source = parent.source
        else:
            self.__source = None

    @property
    def path(self):
        """The path of this builder."""
        s = list()
        c = self
        while c is not None:
            s.append(c.name)
            c = c.parent
        return "/".join(s[::-1])

    @property
    def name(self):
        """The name of this builder."""
        return self.__name

    @property
    def source(self):
        """The source of this builder."""
        return self.__source

    @source.setter
    def source(self, s):
        if self.__source is not None:
            raise AttributeError('Cannot overwrite source.')
        self.__source = s

    @property
    def parent(self):
        """The parent builder of this builder."""
        return self.__parent

    @parent.setter
    def parent(self, p):
        if self.__parent is not None:
            raise AttributeError('Cannot overwrite parent.')
        self.__parent = p
        if self.__source is None:
            self.source = p.source

    def __repr__(self):
        ret = "%s %s %s" % (self.path, self.__class__.__name__, super().__repr__())
        return ret


class BaseBuilder(Builder, metaclass=ABCMeta):
    __attribute = 'attributes'  # self dictionary key for attributes

    @docval({'name': 'name', 'type': str, 'doc': 'The name of the builder.'},
            {'name': 'attributes', 'type': dict, 'doc': 'A dictionary of attributes to create in this builder.',
             'default': dict()},
            {'name': 'parent', 'type': 'GroupBuilder', 'doc': 'The parent builder of this builder.', 'default': None},
            {'name': 'source', 'type': str,
             'doc': 'The source of the data represented in this builder', 'default': None})
    def __init__(self, **kwargs):
        name, attributes, parent, source = getargs('name', 'attributes', 'parent', 'source', kwargs)
        super().__init__(name, parent, source)
        super().__setitem__(BaseBuilder.__attribute, dict())
        for name, val in attributes.items():
            self.set_attribute(name, val)
        self.__location = None

    @property
    def location(self):
        """The location of this Builder in its source."""
        return self.__location

    @location.setter
    def location(self, val):
        self.__location = val

    @property
    def attributes(self):
        """The attributes stored in this Builder object."""
        return super().__getitem__(BaseBuilder.__attribute)

    @docval({'name': 'name', 'type': str, 'doc': 'The name of the attribute.'},
            {'name': 'value', 'type': None, 'doc': 'The attribute value.'})
    def set_attribute(self, **kwargs):
        """Set an attribute for this group."""
        name, value = getargs('name', 'value', kwargs)
        self.attributes[name] = value


class GroupBuilder(BaseBuilder):
    # sub-dictionary keys. subgroups go in super().__getitem__(GroupBuilder.__group)
    __group = 'groups'
    __dataset = 'datasets'
    __link = 'links'
    __attribute = 'attributes'

    @docval({'name': 'name', 'type': str, 'doc': 'The name of the group.'},
            {'name': 'groups', 'type': (dict, list),
             'doc': ('A dictionary or list of subgroups to add to this group. If a dict is provided, only the '
                     'values are used.'),
             'default': dict()},
            {'name': 'datasets', 'type': (dict, list),
             'doc': ('A dictionary or list of datasets to add to this group. If a dict is provided, only the '
                     'values are used.'),
             'default': dict()},
            {'name': 'attributes', 'type': dict, 'doc': 'A dictionary of attributes to create in this group.',
             'default': dict()},
            {'name': 'links', 'type': (dict, list),
             'doc': ('A dictionary or list of links to add to this group. If a dict is provided, only the '
                     'values are used.'),
             'default': dict()},
            {'name': 'parent', 'type': 'GroupBuilder', 'doc': 'The parent builder of this builder.', 'default': None},
            {'name': 'source', 'type': str,
             'doc': 'The source of the data represented in this builder.', 'default': None})
    def __init__(self, **kwargs):
        """Create a builder object for a group."""
        name, groups, datasets, links, attributes, parent, source = getargs(
            'name', 'groups', 'datasets', 'links', 'attributes', 'parent', 'source', kwargs)
        # NOTE: if groups, datasets, or links are dicts, their keys are unused
        groups = self.__to_list(groups)
        datasets = self.__to_list(datasets)
        links = self.__to_list(links)
        # dictionary mapping subgroup/dataset/attribute/link name to the key that maps to the
        # subgroup/dataset/attribute/link sub-dictionary that maps the name to the builder
        self.obj_type = dict()
        super().__init__(name, attributes, parent, source)
        super().__setitem__(GroupBuilder.__group, dict())
        super().__setitem__(GroupBuilder.__dataset, dict())
        super().__setitem__(GroupBuilder.__link, dict())
        for group in groups:
            self.set_group(group)
        for dataset in datasets:
            if dataset is not None:
                self.set_dataset(dataset)
        for link in links:
            self.set_link(link)

    def __to_list(self, d):
        if isinstance(d, dict):
            return list(d.values())
        return d

    @property
    def source(self):
        ''' The source of this Builder '''
        return super().source

    @source.setter
    def source(self, s):
        """Recursively set all subgroups/datasets/links source when this source is set."""
        super(GroupBuilder, self.__class__).source.fset(self, s)
        for group in self.groups.values():
            if group.source is None:
                group.source = s
        for dset in self.datasets.values():
            if dset.source is None:
                dset.source = s
        for link in self.links.values():
            if link.source is None:
                link.source = s

    @property
    def groups(self):
        """The subgroups contained in this group."""
        return super().__getitem__(GroupBuilder.__group)

    @property
    def datasets(self):
        """The datasets contained in this group."""
        return super().__getitem__(GroupBuilder.__dataset)

    @property
    def links(self):
        """The links contained in this group."""
        return super().__getitem__(GroupBuilder.__link)

    @docval(*get_docval(BaseBuilder.set_attribute))
    def set_attribute(self, **kwargs):
        """Set an attribute for this group."""
        name, value = getargs('name', 'value', kwargs)
        self.__check_obj_type(name, GroupBuilder.__attribute)
        super().set_attribute(name, value)
        self.obj_type[name] = GroupBuilder.__attribute

    def __check_obj_type(self, name, obj_type):
        # check that the name is not associated with a different object type in this group
        if name in self.obj_type and self.obj_type[name] != obj_type:
            raise ValueError("'%s' already exists in %s.%s, cannot set in %s."
                             % (name, self.name, self.obj_type[name], obj_type))

    @docval({'name': 'builder', 'type': 'GroupBuilder', 'doc': 'The GroupBuilder to add to this group.'})
    def set_group(self, **kwargs):
        """Add a subgroup to this group."""
        builder = getargs('builder', kwargs)
        self.__set_builder(builder, GroupBuilder.__group)

    @docval({'name': 'builder', 'type': 'DatasetBuilder', 'doc': 'The DatasetBuilder to add to this group.'})
    def set_dataset(self, **kwargs):
        """Add a dataset to this group."""
        builder = getargs('builder', kwargs)
        self.__set_builder(builder, GroupBuilder.__dataset)

    @docval({'name': 'builder', 'type': 'LinkBuilder', 'doc': 'The LinkBuilder to add to this group.'})
    def set_link(self, **kwargs):
        """Add a link to this group."""
        builder = getargs('builder', kwargs)
        self.__set_builder(builder, GroupBuilder.__link)

    def __set_builder(self, builder, obj_type):
        name = builder.name
        self.__check_obj_type(name, obj_type)
        # if child builder already exists (e.g., read from file), do not reset it.
        # resetting the child builder will change the python object ID / hash of the child builder
        # and make the IO backend think that the child builder has not yet been written.
        if self.get(name) == builder:
            return
        super().__getitem__(obj_type)[name] = builder
        self.obj_type[name] = obj_type
        if builder.parent is None:
            builder.parent = self

    def is_empty(self):
        """Returns true if there are no datasets, links, attributes, and non-empty subgroups. False otherwise."""
        if len(self.datasets) or len(self.links) or len(self.attributes):
            return False
        elif len(self.groups):
            return all(g.is_empty() for g in self.groups.values())
        else:
            return True

    def __getitem__(self, key):
        """Like dict.__getitem__, but looks in groups, datasets, attributes, and links sub-dictionaries.
        Key can be a posix path to a sub-builder.
        """
        try:
            key_ar = _posixpath.normpath(key).split('/')
            return self.__get_rec(key_ar)
        except KeyError:
            raise KeyError(key)

    def get(self, key, default=None):
        """Like dict.get, but looks in groups, datasets, attributes, and links sub-dictionaries.
        Key can be a posix path to a sub-builder.
        """
        try:
            key_ar = _posixpath.normpath(key).split('/')
            return self.__get_rec(key_ar)
        except KeyError:
            return default

    def __get_rec(self, key_ar):
        # recursive helper for __getitem__ and get
        if len(key_ar) == 1:
            # get the correct dictionary (groups, datasets, links, attributes) associated with the key
            # then look up the key within that dictionary to get the builder
            return super().__getitem__(self.obj_type[key_ar[0]])[key_ar[0]]
        else:
            if key_ar[0] in self.groups:
                return self.groups[key_ar[0]].__get_rec(key_ar[1:])
        raise KeyError(key_ar[0])

    def __setitem__(self, args, val):
        raise NotImplementedError('__setitem__')

    def __contains__(self, item):
        return self.obj_type.__contains__(item)

    def items(self):
        """Like dict.items, but iterates over items in groups, datasets, attributes, and links sub-dictionaries."""
        return _itertools.chain(self.groups.items(),
                                self.datasets.items(),
                                self.attributes.items(),
                                self.links.items())

    def keys(self):
        """Like dict.keys, but iterates over keys in groups, datasets, attributes, and links sub-dictionaries."""
        return _itertools.chain(self.groups.keys(),
                                self.datasets.keys(),
                                self.attributes.keys(),
                                self.links.keys())

    def values(self):
        """Like dict.values, but iterates over values in groups, datasets, attributes, and links sub-dictionaries."""
        return _itertools.chain(self.groups.values(),
                                self.datasets.values(),
                                self.attributes.values(),
                                self.links.values())


class DatasetBuilder(BaseBuilder):
    OBJECT_REF_TYPE = 'object'
    REGION_REF_TYPE = 'region'

    @docval({'name': 'name', 'type': str, 'doc': 'The name of the dataset.'},
            {'name': 'data',
             'type': ('array_data', 'scalar_data', 'data', 'DatasetBuilder', 'RegionBuilder', Iterable, datetime),
             'doc': 'The data in this dataset.', 'default': None},
            {'name': 'dtype', 'type': (type, np.dtype, str, list),
             'doc': 'The datatype of this dataset.', 'default': None},
            {'name': 'attributes', 'type': dict,
             'doc': 'A dictionary of attributes to create in this dataset.', 'default': dict()},
            {'name': 'maxshape', 'type': (int, tuple),
             'doc': 'The shape of this dataset. Use None for scalars.', 'default': None},
            {'name': 'chunks', 'type': bool, 'doc': 'Whether or not to chunk this dataset.', 'default': False},
            {'name': 'parent', 'type': GroupBuilder, 'doc': 'The parent builder of this builder.', 'default': None},
            {'name': 'source', 'type': str, 'doc': 'The source of the data in this builder.', 'default': None})
    def __init__(self, **kwargs):
        """ Create a Builder object for a dataset """
        name, data, dtype, attributes, maxshape, chunks, parent, source = getargs(
            'name', 'data', 'dtype', 'attributes', 'maxshape', 'chunks', 'parent', 'source', kwargs)
        super().__init__(name, attributes, parent, source)
        self['data'] = data
        self['attributes'] = _copy.copy(attributes)
        self.__chunks = chunks
        self.__maxshape = maxshape
        if isinstance(data, BaseBuilder):
            if dtype is None:
                dtype = self.OBJECT_REF_TYPE
        self.__dtype = dtype
        self.__name = name

    @property
    def data(self):
        """The data stored in the dataset represented by this builder."""
        return self['data']

    @data.setter
    def data(self, val):
        if self['data'] is not None:
            raise AttributeError("Cannot overwrite data.")
        self['data'] = val

    @property
    def chunks(self):
        """Whether or not this dataset is chunked."""
        return self.__chunks

    @property
    def maxshape(self):
        """The max shape of this dataset."""
        return self.__maxshape

    @property
    def dtype(self):
        """The data type of this dataset."""
        return self.__dtype

    @dtype.setter
    def dtype(self, val):
        if self.__dtype is not None:
            raise AttributeError("Cannot overwrite dtype.")
        self.__dtype = val


class LinkBuilder(Builder):

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder),
             'doc': 'The target group or dataset of this link.'},
            {'name': 'name', 'type': str, 'doc': 'The name of the link', 'default': None},
            {'name': 'parent', 'type': GroupBuilder, 'doc': 'The parent builder of this builder', 'default': None},
            {'name': 'source', 'type': str, 'doc': 'The source of the data in this builder', 'default': None})
    def __init__(self, **kwargs):
        """Create a builder object for a link."""
        name, builder, parent, source = getargs('name', 'builder', 'parent', 'source', kwargs)
        if name is None:
            name = builder.name
        super().__init__(name, parent, source)
        self['builder'] = builder

    @property
    def builder(self):
        """The target builder object."""
        return self['builder']


class ReferenceBuilder(dict):

    @docval({'name': 'builder', 'type': (DatasetBuilder, GroupBuilder),
             'doc': 'The group or dataset this reference applies to.'})
    def __init__(self, **kwargs):
        """Create a builder object for a reference."""
        builder = getargs('builder', kwargs)
        self['builder'] = builder

    @property
    def builder(self):
        """The target builder object."""
        return self['builder']


class RegionBuilder(ReferenceBuilder):

    @docval({'name': 'region', 'type': (slice, tuple, list, RegionReference),
             'doc': 'The region, i.e. slice or indices, into the target dataset.'},
            {'name': 'builder', 'type': DatasetBuilder, 'doc': 'The dataset this region reference applies to.'})
    def __init__(self, **kwargs):
        """Create a builder object for a region reference."""
        region, builder = getargs('region', 'builder', kwargs)
        super().__init__(builder)
        self['region'] = region

    @property
    def region(self):
        """The selected region of the target dataset."""
        return self['region']



from copy import deepcopy
from datetime import datetime

import numpy as np

from ..container import Container, Data, DataRegion, MultiContainerInterface
from ..spec import AttributeSpec, LinkSpec, RefSpec, GroupSpec
from ..spec.spec import BaseStorageSpec, ZERO_OR_MANY, ONE_OR_MANY
from ..utils import docval, getargs, ExtenderMeta, get_docval, popargs, AllowPositional


class ClassGenerator:

    def __init__(self):
        self.__custom_generators = []

    @property
    def custom_generators(self):
        return self.__custom_generators

    @docval({'name': 'generator', 'type': type, 'doc': 'the CustomClassGenerator class to register'})
    def register_generator(self, **kwargs):
        """Add a custom class generator to this ClassGenerator.

        Generators added later are run first. Duplicates are moved to the top of the list.
        """
        generator = getargs('generator', kwargs)
        if not issubclass(generator, CustomClassGenerator):
            raise ValueError('Generator %s must be a subclass of CustomClassGenerator.' % generator)
        if generator in self.__custom_generators:
            self.__custom_generators.remove(generator)
        self.__custom_generators.insert(0, generator)

    @docval({'name': 'data_type', 'type': str, 'doc': 'the data type to create a AbstractContainer class for'},
            {'name': 'spec', 'type': BaseStorageSpec, 'doc': ''},
            {'name': 'parent_cls', 'type': type, 'doc': ''},
            {'name': 'attr_names', 'type': dict, 'doc': ''},
            {'name': 'type_map', 'type': 'TypeMap', 'doc': ''},
            returns='the class for the given namespace and data_type', rtype=type)
    def generate_class(self, **kwargs):
        """Get the container class from data type specification.
        If no class has been associated with the ``data_type`` from ``namespace``, a class will be dynamically
        created and returned.
        """
        data_type, spec, parent_cls, attr_names, type_map = getargs('data_type', 'spec', 'parent_cls', 'attr_names',
                                                                    'type_map', kwargs)

        not_inherited_fields = dict()
        for k, field_spec in attr_names.items():
            if k == 'help':  # pragma: no cover
                # (legacy) do not add field named 'help' to any part of class object
                continue
            if isinstance(field_spec, GroupSpec) and field_spec.data_type is None:  # skip named, untyped groups
                continue
            if not spec.is_inherited_spec(field_spec):
                not_inherited_fields[k] = field_spec
        try:
            classdict = dict()
            bases = [parent_cls]
            docval_args = list(deepcopy(get_docval(parent_cls.__init__)))
            for attr_name, field_spec in not_inherited_fields.items():
                for class_generator in self.__custom_generators:  # pragma: no branch
                    # each generator can update classdict and docval_args
                    if class_generator.apply_generator_to_field(field_spec, bases, type_map):
                        class_generator.process_field_spec(classdict, docval_args, parent_cls, attr_name,
                                                           not_inherited_fields, type_map, spec)
                        break  # each field_spec should be processed by only one generator

            for class_generator in self.__custom_generators:
                class_generator.post_process(classdict, bases, docval_args, spec)

            for class_generator in reversed(self.__custom_generators):
                # go in reverse order so that base init is added first and
                # later class generators can modify or overwrite __init__ set by an earlier class generator
                class_generator.set_init(classdict, bases, docval_args, not_inherited_fields, spec.name)
        except TypeDoesNotExistError as e:  # pragma: no cover
            # this error should never happen after hdmf#322
            name = spec.data_type_def
            if name is None:
                name = 'Unknown'
            raise ValueError("Cannot dynamically generate class for type '%s'. " % name
                             + str(e)
                             + " Please define that type before defining '%s'." % name)
        cls = ExtenderMeta(data_type, tuple(bases), classdict)
        return cls


class TypeDoesNotExistError(Exception):  # pragma: no cover
    pass


class CustomClassGenerator:
    """Subclass this class and register an instance to alter how classes are auto-generated."""

    def __new__(cls, *args, **kwargs):  # pragma: no cover
        raise TypeError('Cannot instantiate class %s' % cls.__name__)

    # mapping from spec types to allowable python types for docval for fields during dynamic class generation
    # e.g., if a dataset/attribute spec has dtype int32, then get_class should generate a docval for the class'
    # __init__ method that allows the types (int, np.int32, np.int64) for the corresponding field.
    # passing an np.int16 would raise a docval error.
    # passing an int64 to __init__ would result in the field storing the value as an int64 (and subsequently written
    # as an int64). no upconversion or downconversion happens as a result of this map
    _spec_dtype_map = {
        'float32': (float, np.float32, np.float64),
        'float': (float, np.float32, np.float64),
        'float64': (float, np.float64),
        'double': (float, np.float64),
        'int8': (np.int8, np.int16, np.int32, np.int64, int),
        'int16': (np.int16, np.int32, np.int64, int),
        'short': (np.int16, np.int32, np.int64, int),
        'int32': (int, np.int32, np.int64),
        'int': (int, np.int32, np.int64),
        'int64': np.int64,
        'long': np.int64,
        'uint8': (np.uint8, np.uint16, np.uint32, np.uint64),
        'uint16': (np.uint16, np.uint32, np.uint64),
        'uint32': (np.uint32, np.uint64),
        'uint64': np.uint64,
        'numeric': (float, np.float32, np.float64, np.int8, np.int16, np.int32, np.int64, int, np.uint8, np.uint16,
                    np.uint32, np.uint64),
        'text': str,
        'utf': str,
        'utf8': str,
        'utf-8': str,
        'ascii': bytes,
        'bytes': bytes,
        'bool': (bool, np.bool_),
        'isodatetime': datetime,
        'datetime': datetime
    }

    @classmethod
    def _get_type_from_spec_dtype(cls, spec_dtype):
        """Get the Python type associated with the given spec dtype string.
        Raises ValueError if the given dtype has no mapping to a Python type.
        """
        dtype = cls._spec_dtype_map.get(spec_dtype)
        if dtype is None:  # pragma: no cover
            # this should not happen as long as _spec_dtype_map is kept up to date with
            # hdmf.spec.spec.DtypeHelper.valid_primary_dtypes
            raise ValueError("Spec dtype '%s' cannot be mapped to a Python type." % spec_dtype)
        return dtype

    @classmethod
    def _get_container_type(cls, type_name, type_map):
        """Search all namespaces for the container class associated with the given data type.
        Raises TypeDoesNotExistError if type is not found in any namespace.
        """
        container_type = type_map.get_dt_container_cls(type_name)
        if container_type is None:  # pragma: no cover
            # this should never happen after hdmf#322
            raise TypeDoesNotExistError("Type '%s' does not exist." % type_name)
        return container_type

    @classmethod
    def _get_type(cls, spec, type_map):
        """Get the type of a spec for use in docval.
        Returns a container class, a type, a tuple of types, ('array_data', 'data') for specs with
        non-scalar shape, or (Data, Container) when an attribute reference target has not been mapped to a container
        class.
        """
        if isinstance(spec, AttributeSpec):
            if isinstance(spec.dtype, RefSpec):
                try:
                    container_type = cls._get_container_type(spec.dtype.target_type, type_map)
                    return container_type
                except TypeDoesNotExistError:
                    # TODO what happens when the attribute ref target is not (or not yet) mapped to a container class?
                    # returning Data, Container works as a generic fallback for now but should be more specific
                    return Data, Container
            elif spec.shape is None and spec.dims is None:
                return cls._get_type_from_spec_dtype(spec.dtype)
            else:
                return 'array_data', 'data'
        if isinstance(spec, LinkSpec):
            return cls._get_container_type(spec.target_type, type_map)
        if spec.data_type is not None:
            return cls._get_container_type(spec.data_type, type_map)
        if spec.shape is None and spec.dims is None:
            return cls._get_type_from_spec_dtype(spec.dtype)
        return 'array_data', 'data'

    @classmethod
    def _ischild(cls, dtype):
        """Check if dtype represents a type that is a child."""
        ret = False
        if isinstance(dtype, tuple):
            for sub in dtype:
                ret = ret or cls._ischild(sub)
        elif isinstance(dtype, type) and issubclass(dtype, (Container, Data, DataRegion)):
            ret = True
        return ret

    @staticmethod
    def _set_default_name(docval_args, default_name):
        """Set the default value for the name docval argument."""
        if default_name is not None:
            for x in docval_args:
                if x['name'] == 'name':
                    x['default'] = default_name

    @classmethod
    def apply_generator_to_field(cls, field_spec, bases, type_map):
        """Return True to signal that this generator should return on all fields not yet processed."""
        return True

    @classmethod
    def process_field_spec(cls, classdict, docval_args, parent_cls, attr_name, not_inherited_fields, type_map, spec):
        """Add __fields__ to the classdict and update the docval args for the field spec with the given attribute name.
        :param classdict: The dict to update with __fields__ (or a different parent_cls._fieldsname).
        :param docval_args: The list of docval arguments.
        :param parent_cls: The parent class.
        :param attr_name: The attribute name of the field spec for the container class to generate.
        :param not_inherited_fields: Dictionary of fields not inherited from the parent class.
        :param type_map: The type map to use.
        :param spec: The spec for the container class to generate.
        """
        field_spec = not_inherited_fields[attr_name]
        dtype = cls._get_type(field_spec, type_map)
        fields_conf = {'name': attr_name,
                       'doc': field_spec['doc']}
        if cls._ischild(dtype) and issubclass(parent_cls, Container) and not isinstance(field_spec, LinkSpec):
            fields_conf['child'] = True
        # if getattr(field_spec, 'value', None) is not None:  # TODO set the fixed value on the class?
        #     fields_conf['settable'] = False
        classdict.setdefault(parent_cls._fieldsname, list()).append(fields_conf)

        docval_arg = dict(
            name=attr_name,
            doc=field_spec.doc,
            type=cls._get_type(field_spec, type_map)
        )
        shape = getattr(field_spec, 'shape', None)
        if shape is not None:
            docval_arg['shape'] = shape
        if cls._check_spec_optional(field_spec, spec):
            docval_arg['default'] = getattr(field_spec, 'default_value', None)
        cls._add_to_docval_args(docval_args, docval_arg)

    @classmethod
    def _check_spec_optional(cls, field_spec, spec):
        """Returns True if the spec or any of its parents (up to the parent type spec) are optional."""
        if not field_spec.required:
            return True
        if field_spec == spec:
            return False
        if field_spec.parent is not None:
            return cls._check_spec_optional(field_spec.parent, spec)

    @classmethod
    def _add_to_docval_args(cls, docval_args, arg, err_if_present=False):
        """Add the docval arg to the list if not present. If present, overwrite it in place or raise an error."""
        inserted = False
        for i, x in enumerate(docval_args):
            if x['name'] == arg['name']:
                if err_if_present:
                    raise ValueError("Argument %s already exists in docval args." % arg["name"])
                docval_args[i] = arg
                inserted = True
        if not inserted:
            docval_args.append(arg)

    @classmethod
    def post_process(cls, classdict, bases, docval_args, spec):
        """Convert classdict['__fields__'] to tuple and update docval args for a fixed name and default name.
        :param classdict: The class dictionary to convert with '__fields__' key (or a different bases[0]._fieldsname)
        :param bases: The list of base classes.
        :param docval_args: The dict of docval arguments.
        :param spec: The spec for the container class to generate.
        """
        # convert classdict['__fields__'] from list to tuple if present
        for b in bases:
            fields = classdict.get(b._fieldsname)
            if fields is not None and not isinstance(fields, tuple):
                classdict[b._fieldsname] = tuple(fields)

        # if spec provides a fixed name for this type, remove the 'name' arg from docval_args so that values cannot
        # be passed for a name positional or keyword arg
        if spec.name is not None:
            for arg in list(docval_args):
                if arg['name'] == 'name':
                    docval_args.remove(arg)

        # set default name in docval args if provided
        cls._set_default_name(docval_args, spec.default_name)

    @classmethod
    def set_init(cls, classdict, bases, docval_args, not_inherited_fields, name):
        # get docval arg names from superclass
        base = bases[0]
        parent_docval_args = set(arg['name'] for arg in get_docval(base.__init__))
        new_args = list()
        for attr_name, field_spec in not_inherited_fields.items():
            # store arguments for fields that are not in the superclass and not in the superclass __init__ docval
            # so that they are set after calling base.__init__
            if attr_name not in parent_docval_args:
                new_args.append(attr_name)

        @docval(*docval_args, allow_positional=AllowPositional.WARNING)
        def __init__(self, **kwargs):
            if name is not None:  # force container name to be the fixed name in the spec
                kwargs.update(name=name)

            # remove arguments from kwargs that correspond to fields that are new (not inherited)
            # set these arguments after calling base.__init__
            new_kwargs = dict()
            for f in new_args:
                new_kwargs[f] = popargs(f, kwargs) if f in kwargs else None

            # NOTE: the docval of some constructors do not include all of the fields. the constructor may set
            # some fields to fixed values. so only keep the kwargs that are used in the constructor docval
            kwargs_to_pass = {k: v for k, v in kwargs.items() if k in parent_docval_args}

            base.__init__(self, **kwargs_to_pass)  # special case: need to pass self to __init__
            # TODO should super() be used above instead of base?

            # set the fields that are new to this class (not inherited)
            for f, arg_val in new_kwargs.items():
                setattr(self, f, arg_val)

        classdict['__init__'] = __init__


class MCIClassGenerator(CustomClassGenerator):

    @classmethod
    def apply_generator_to_field(cls, field_spec, bases, type_map):
        """Return True if the field spec has quantity * or +, False otherwise."""
        return getattr(field_spec, 'quantity', None) in (ZERO_OR_MANY, ONE_OR_MANY)

    @classmethod
    def process_field_spec(cls, classdict, docval_args, parent_cls, attr_name, not_inherited_fields, type_map, spec):
        """Add __clsconf__ to the classdict and update the docval args for the field spec with the given attribute name.
        :param classdict: The dict to update with __clsconf__.
        :param docval_args: The list of docval arguments.
        :param parent_cls: The parent class.
        :param attr_name: The attribute name of the field spec for the container class to generate.
        :param not_inherited_fields: Dictionary of fields not inherited from the parent class.
        :param type_map: The type map to use.
        :param spec: The spec for the container class to generate.
        """
        field_spec = not_inherited_fields[attr_name]
        field_clsconf = dict(
            attr=attr_name,
            type=cls._get_type(field_spec, type_map),
            add='add_{}'.format(attr_name),
            get='get_{}'.format(attr_name),
            create='create_{}'.format(attr_name)
        )
        classdict.setdefault('__clsconf__', list()).append(field_clsconf)

        # add a specialized docval arg for __init__
        docval_arg = dict(
            name=attr_name,
            doc=field_spec.doc,
            type=(list, tuple, dict, cls._get_type(field_spec, type_map))
        )
        if cls._check_spec_optional(field_spec, spec):
            docval_arg['default'] = getattr(field_spec, 'default_value', None)
        cls._add_to_docval_args(docval_args, docval_arg)

    @classmethod
    def post_process(cls, classdict, bases, docval_args, spec):
        """Add MultiContainerInterface to the list of base classes.
        :param classdict: The class dictionary.
        :param bases: The list of base classes.
        :param docval_args: The dict of docval arguments.
        :param spec: The spec for the container class to generate.
        """
        if '__clsconf__' in classdict:
            # do not add MCI as a base if a base is already a subclass of MultiContainerInterface
            for b in bases:
                if issubclass(b, MultiContainerInterface):
                    break
            else:
                if issubclass(MultiContainerInterface, bases[0]):
                    # if bases[0] is Container or another superclass of MCI, then make sure MCI goes first
                    # otherwise, MRO is ambiguous
                    bases.insert(0, MultiContainerInterface)
                else:
                    # bases[0] is not a subclass of MCI and not a superclass of MCI. place that class first
                    # before MCI. that class __init__ should call super().__init__ which will call the
                    # MCI init
                    bases.insert(1, MultiContainerInterface)

    @classmethod
    def set_init(cls, classdict, bases, docval_args, not_inherited_fields, name):
        if '__clsconf__' in classdict:
            previous_init = classdict['__init__']

            @docval(*docval_args, allow_positional=AllowPositional.WARNING)
            def __init__(self, **kwargs):
                # store the values passed to init for each MCI attribute so that they can be added
                # after calling __init__
                new_kwargs = list()
                for field_clsconf in classdict['__clsconf__']:
                    attr_name = field_clsconf['attr']
                    # do not store the value if it is None or not present
                    if attr_name not in kwargs or kwargs[attr_name] is None:
                        continue
                    add_method_name = field_clsconf['add']
                    new_kwarg = dict(
                        attr_name=attr_name,
                        value=popargs(attr_name, kwargs),
                        add_method_name=add_method_name
                    )
                    new_kwargs.append(new_kwarg)

                # call the parent class init without the MCI attribute
                previous_init(self, **kwargs)

                # call the add method for each MCI attribute
                for new_kwarg in new_kwargs:
                    add_method = getattr(self, new_kwarg['add_method_name'])
                    add_method(new_kwarg['value'])

            # override __init__
            classdict['__init__'] = __init__
